import io
import pathlib
import tempfile
import pytest
from threatexchange.cli.tests.e2e_test_helper import (
    ThreatExchangeCLIE2eHelper,
    te_cli,
)


@pytest.fixture
def hash_cli(
    te_cli: ThreatExchangeCLIE2eHelper,
) -> ThreatExchangeCLIE2eHelper:
    te_cli.COMMON_CALL_ARGS = ("hash",)
    return te_cli


@pytest.fixture
def tmp_file():
    with tempfile.NamedTemporaryFile() as f:
        yield pathlib.Path(f.name)


def test_file(hash_cli: ThreatExchangeCLIE2eHelper, tmp_file: pathlib.Path):
    tmp_file.write_text("http://evil.com")
    hash_cli.assert_cli_output(
        ("url", str(tmp_file)),
        "url_md5 6d3af727a4e7b025fd59a5469b3a9c57",
    )

    hash_cli.assert_cli_usage_error(("url", "blah.txt"))


def test_dashdash(hash_cli: ThreatExchangeCLIE2eHelper):
    hash_cli.assert_cli_output(
        ("url", "--", "http://evil.com"),
        "url_md5 6d3af727a4e7b025fd59a5469b3a9c57",
    )


def test_stdin(
    hash_cli: ThreatExchangeCLIE2eHelper,
    monkeypatch: pytest.MonkeyPatch,
    tmp_file: pathlib.Path,
):
    tmp_file.write_text("http://evil.com")

    with tmp_file.open() as fake_stin:
        monkeypatch.setattr("sys.stdin", fake_stin)
        hash_cli.assert_cli_output(
            ("url", "-"),
            "url_md5 6d3af727a4e7b025fd59a5469b3a9c57",
        )


def test_mixed(hash_cli: ThreatExchangeCLIE2eHelper, tmp_path: pathlib.Path):
    a = tmp_path / "a"
    b = tmp_path / "b"
    a.write_text("http://evil.com", None)
    b.touch()

    hash_cli.assert_cli_output(
        ("url", str(a), str(b), "--", "fb.com"),
        [
            "url_md5 6d3af727a4e7b025fd59a5469b3a9c57",
            "url_md5 d41d8cd98f00b204e9800998ecf8427e",
            "url_md5 fb8191ebebc85f9eb6fd21e198f20979",
        ],
    )


def test_missing(hash_cli: ThreatExchangeCLIE2eHelper):
    hash_cli.assert_cli_usage_error(())
    hash_cli.assert_cli_usage_error(("photo",))



import pytest
from threatexchange.cli.tests.e2e_test_helper import (
    ThreatExchangeCLIE2eHelper,
    te_cli,
)
from threatexchange.exchanges.clients.ncmec.hash_api import NCMECEnvironment


@pytest.fixture
def cli(
    te_cli: ThreatExchangeCLIE2eHelper,
) -> ThreatExchangeCLIE2eHelper:
    te_cli.COMMON_CALL_ARGS = ("config", "collab")
    return te_cli


def test_required_argument(cli: ThreatExchangeCLIE2eHelper) -> None:
    name = "A config name"
    cli.assert_cli_usage_error(("edit", "ncmec", "-C", name))
    cli.cli_call(
        "edit", "ncmec", "-C", name, f"--environment={NCMECEnvironment.test_NGO.name}"
    )
    cli.assert_cli_output((), expected_output=f"ncmec {name}")  # Defaults to list
    cli.assert_cli_output(("list",), expected_output=f"ncmec {name}")


def test_editing_collab(cli: ThreatExchangeCLIE2eHelper) -> None:
    name = "collab_name"
    expected = """
{
  "environment": "%s",
  "name": "collab_name",
  "api": "ncmec",
  "enabled": true,
  "only_esp_ids": []
}
    """.strip()
    cli.cli_call(
        "edit",
        "ncmec",
        "-C",
        name,
        f"--environment={NCMECEnvironment.test_NGO.name}",
    )
    cli.assert_cli_output(["print", name], expected % NCMECEnvironment.test_NGO.value)
    cli.cli_call(
        "edit",
        "ncmec",
        name,
        f"--environment={NCMECEnvironment.test_Exploitative.name}",
    )

    cli.assert_cli_output(
        ["print", name], expected % NCMECEnvironment.test_Exploitative.value
    )


def test_complex_types(cli: ThreatExchangeCLIE2eHelper) -> None:
    cli.cli_call(
        "tx",
        "config",
        "extensions",
        "add",
        "threatexchange.cli.tests.fake_extension",
    )
    cli.cli_call(
        "edit",
        "fake",
        "fake",
        "--an-int=1",
        "--a-str=a",
        "--an-enum=OPTION_A",
        "--a-list=1,2,3",
        "--a-set=3,2",
        "--create",
    )
    expected = """
{
  "an_int": 1,
  "a_str": "a",
  "a_list": [
    "1",
    "2",
    "3"
  ],
  "a_set": [
    2,
    3
  ],
  "an_enum": "a",
  "name": "fake",
  "api": "fake",
  "enabled": true,
  "optional": null
}
    """.strip()
    cli.assert_cli_output(["print", "fake"], expected)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper for the `threatexchange` library, can serve as a simple e2e solution.

The flow the CLI is generally:
  1. Configure collaborations and APIs with (or skip to use fake sample data)
     $ threatexchange config collab edit
     $ threatexchange config api  # to set up credentials if needed
  2. Fetch data and build match indices
     $ threatexchange fetch
  3. Match data
     $ threatexchange match photo my_photo.jpg
  4. Contribute labels to external APIs
     $ threatexchange label photo my_photo.jpg dog 
     $ threatexchange label photo my_photo.jpg --false-positive

Additionally, there are a number of utility commands:
  * threatexchange dataset
  * threatexchange hash

See the --help of subcommands for more information.

State is persisted between runs, entirely in the ~/.threatexchange directory
"""

import argparse
from contextlib import contextmanager
from importlib import metadata
import logging
import inspect
import os
import subprocess
import sys
import typing as t
import pathlib
import shutil
import warnings

# Import pdq first with its hash order warning squelched, it's before our time
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    from threatexchange.signal_type.pdq import signal

from threatexchange.cli.exceptions import CommandError
from threatexchange import interface_validation
from threatexchange.content_type.content_base import ContentType
from threatexchange.extensions.manifest import ThreatExchangeExtensionManifest
from threatexchange.exchanges.impl.file_api import LocalFileSignalExchangeAPI
from threatexchange.exchanges.impl.static_sample import StaticSampleSignalExchangeAPI
from threatexchange.exchanges.impl.fb_threatexchange_api import (
    FBThreatExchangeCredentials,
    FBThreatExchangeSignalExchangeAPI,
)
from threatexchange.exchanges.impl.stop_ncii_api import (
    StopNCIICredentials,
    StopNCIISignalExchangeAPI,
)
from threatexchange.exchanges.impl.ncmec_api import (
    NCMECCredentials,
    NCMECSignalExchangeAPI,
)

from threatexchange.content_type import photo, video, text, url
from threatexchange.exchanges.signal_exchange_api import SignalExchangeAPI
from threatexchange.exchanges.auth import (
    SignalExchangeAPIInvalidAuthException,
    SignalExchangeAPIMissingAuthException,
)
from threatexchange.signal_type import (
    md5,
    raw_text,
    url as url_signal,
    url_md5,
    trend_query,
)
from threatexchange.cli.cli_config import CLiConfig, CliState
from threatexchange.cli.cli_config import CLISettings
from threatexchange.cli import (
    command_base as base,
    fetch_cmd,
    label_cmd,
    dataset_cmd,
    hash_cmd,
    match_cmd,
    config_cmd,
)
from threatexchange.signal_type.signal_base import SignalType

_DEFAULT_SIGNAL_TYPES: t.Sequence[t.Type[SignalType]] = [
    signal.PdqSignal,
    md5.VideoMD5Signal,
    raw_text.RawTextSignal,
    url_signal.URLSignal,
    url_md5.UrlMD5Signal,
    trend_query.TrendQuerySignal,
]


def get_subcommands() -> t.List[t.Type[base.Command]]:
    return [
        config_cmd.ConfigCommand,
        fetch_cmd.FetchCommand,
        match_cmd.MatchCommand,
        label_cmd.LabelCommand,
        dataset_cmd.DatasetCommand,
        hash_cmd.HashCommand,
    ]


def get_argparse(settings: CLISettings) -> argparse.ArgumentParser:
    ap = argparse.ArgumentParser(
        description=__doc__, formatter_class=argparse.RawDescriptionHelpFormatter
    )
    # is_config = Safe mode to try and let you fix bad settings from CLI
    ap.set_defaults(is_config=False)
    ap.add_argument(
        "--factory-reset",
        action="store_true",
        help="Remove all state, bringing you back to a fresh install",
    )
    ap.add_argument(
        "--version",
        action="store_true",
        dest="print_version",
        help="print the version and exit",
    )
    ap.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=0,
        help="display logging output (repeat for more)",
    )
    subparsers = ap.add_subparsers(
        dest="toplevel_command_name", title="verbs", help="which action to do"
    )
    for command in get_subcommands():
        command.add_command_to_subparser(settings, subparsers)

    return ap


def execute_command(settings: CLISettings, namespace) -> None:
    assert hasattr(namespace, "command_cls")
    command_cls: t.Type[base.Command] = namespace.command_cls
    logging.debug("Setup complete, handing off to %s", command_cls.__name__)
    # Init everything
    command_argspec = inspect.getfullargspec(command_cls.__init__)
    arg_names = set(command_argspec[0])
    # Since we didn't import click, use hard-to-debug magic to init the command
    command_args = {k: v for k, v in namespace.__dict__.items() if k in arg_names}
    if "full_argparse_namespace" in arg_names:
        command_args["full_argparse_namespace"] = namespace

    command = command_cls(**command_args)
    command.execute(settings)


@contextmanager
def _handle_api_creds(config: CLiConfig) -> t.Iterator[None]:
    te_creds = None
    ncmec_creds = None
    stop_ncii_creds = config.stop_ncii_keys

    def cfg_cmd(src_api: t.Type[SignalExchangeAPI], flags: str) -> str:
        return f"threatexchange config api {src_api.get_name()} {flags}"

    if config.fb_threatexchange_api_token:
        te_creds = FBThreatExchangeCredentials(config.fb_threatexchange_api_token)
    if config.ncmec_credentials:
        ncmec_creds = NCMECCredentials(*config.ncmec_credentials)

    with FBThreatExchangeCredentials.set_default(
        te_creds, cfg_cmd(FBThreatExchangeSignalExchangeAPI, "--api-token")
    ), NCMECCredentials.set_default(
        ncmec_creds, cfg_cmd(NCMECSignalExchangeAPI, "--user --pass")
    ), StopNCIICredentials.set_default(
        stop_ncii_creds, cfg_cmd(StopNCIISignalExchangeAPI, "<TBD>")  # TODO
    ):
        try:
            yield
        except SignalExchangeAPIInvalidAuthException as ia:
            logging.exception("Original invalid auth error")
            raise CommandError.user(
                f"Invalid auth for {ia.src_api.get_name()}: {ia.message}"
            ) from ia
        except SignalExchangeAPIMissingAuthException as ma:
            logging.exception("Original missing auth error")
            raise CommandError.user(ma.pretty_str()) from ma


class _ExtendedTypes(t.NamedTuple):
    content_types: t.List[t.Type[ContentType]]
    signal_types: t.List[t.Type[SignalType]]
    api_types: t.List[t.Type[SignalExchangeAPI]]
    load_failures: t.List[str]

    def assert_no_errors(self) -> None:
        if not self.load_failures:
            return
        err_list = "\n  ".join(self.load_failures)
        raise base.CommandError.user(
            "Some extensions are no longer loadable! You might need to "
            "re-install, or else remove them with the "
            "`threatexchange config extensions remove` command:\n  "
            f"{err_list}"
        )


def _get_extended_functionality(config: CLiConfig) -> _ExtendedTypes:
    ret = _ExtendedTypes([], [], [], [])
    for extension in config.extensions:
        logging.debug("Loading extension %s", extension)
        try:
            manifest = ThreatExchangeExtensionManifest.load_from_module_name(extension)
        except (ValueError, ImportError):
            ret.load_failures.append(extension)
        else:
            ret.signal_types.extend(manifest.signal_types)
            ret.content_types.extend(manifest.content_types)
            ret.api_types.extend(manifest.apis)
    return ret


def _get_settings(
    config: CLiConfig, dir: pathlib.Path
) -> t.Tuple[CLISettings, _ExtendedTypes]:
    """
    Configure the behavior and functionality.
    """

    extensions = _get_extended_functionality(config)

    signals = interface_validation.SignalTypeMapping(
        [photo.PhotoContent, video.VideoContent, url.URLContent, text.TextContent]
        + extensions.content_types,
        list(_DEFAULT_SIGNAL_TYPES) + extensions.signal_types,
    )
    base_apis: t.List[t.Type[SignalExchangeAPI]] = [
        StaticSampleSignalExchangeAPI,
        LocalFileSignalExchangeAPI,
        StopNCIISignalExchangeAPI,
        NCMECSignalExchangeAPI,
        FBThreatExchangeSignalExchangeAPI,
    ]
    apis = interface_validation.SignalExchangeAPIMapping(
        base_apis + extensions.api_types
    )
    state = CliState(list(apis.api_by_name.values()), dir=dir)

    return (
        CLISettings(
            interface_validation.FunctionalityMapping(signals, apis, state), state
        ),
        extensions,
    )


def _setup_logging(level_str: str, *, initial: bool = False) -> None:
    level = logging.DEBUG
    if level_str == "0":
        level = logging.CRITICAL
    elif level_str == "1":
        level = logging.INFO
    if initial:
        logging.basicConfig(
            format="%(asctime)s %(levelname).1s] %(message)s", level=level, force=True
        )
    else:
        logging.getLogger().setLevel(level)


def inner_main(
    args: t.Optional[t.Sequence[t.Text]] = None,
    state_dir: pathlib.Path = pathlib.Path("~/.threatexchange"),
) -> None:
    """The main called by tests"""
    config = CliState(
        [], state_dir
    ).get_persistent_config()  # TODO fix the circular dependency
    settings, extensions = _get_settings(config, state_dir)
    ap = get_argparse(settings)
    namespace = ap.parse_args(args)
    if namespace.verbose:
        _setup_logging(str(namespace.verbose))
    if namespace.factory_reset:
        print("Resetting to factory defaults.", file=sys.stderr)
        shutil.rmtree(str(state_dir.expanduser().absolute()))
        return
    if namespace.print_version:
        here = pathlib.Path(os.path.realpath(__file__)).parent
        if (here / ".local_install_detection.txt").is_file():
            version_txt = (here.parent.parent / "version.txt").read_text().strip()
            git_hash = subprocess.run(
                ["git", "rev-parse", "HEAD"],
                cwd=str(here.absolute()),
                capture_output=True,
                text=True,
            ).stdout.strip()
            version = f"{version_txt or '???'} dev-{git_hash or '???'}"
        else:
            version = metadata.version("threatexchange")
        print(version)
        return
    if not namespace.is_config:
        extensions.assert_no_errors()
    if not namespace.toplevel_command_name:
        ap.print_help()
        return
    with _handle_api_creds(settings.get_persistent_config()):
        execute_command(settings, namespace)


def main():
    """The main called by pip"""
    _setup_logging(os.getenv("TX_VERBOSE", "0"), initial=True)
    try:
        inner_main()
    except base.CommandError as ce:
        print(ce, file=sys.stderr)
        sys.exit(ce.returncode)
    except KeyboardInterrupt:
        # No stack for CTRL+C
        sys.exit(130)


# Surprise! This line is not actually called when installed as a script by pip
if __name__ == "__main__":
    main()  # Don't add anything else here



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
SignalExchangeAPI impl for Facebook/Meta's ThreatExchange Graph API platform.

https://developers.facebook.com/programs/threatexchange
https://developers.facebook.com/docs/threat-exchange/reference/apis/
"""


from collections import defaultdict
import typing as t
import time
from dataclasses import dataclass, field
from threatexchange.exchanges.clients.fb_threatexchange.threat_updates import (
    ThreatUpdateJSON,
)

from threatexchange.exchanges.clients.fb_threatexchange.api import (
    ThreatExchangeAPI,
    is_valid_app_token,
)

from threatexchange.exchanges import auth, fetch_state as state
from threatexchange.exchanges.signal_exchange_api import (
    SignalExchangeAPIWithSimpleUpdates,
)
from threatexchange.exchanges.collab_config import CollaborationConfigWithDefaults
from threatexchange.signal_type.signal_base import SignalType
from threatexchange.exchanges.impl.fb_threatexchange_signal import (
    HasFbThreatExchangeIndicatorType,
)

_API_NAME = "fb_threatexchange"


@dataclass
class _FBThreatExchangeCollabConfigRequiredFields:
    privacy_group: int = field(
        metadata={
            "help": "ThreatPrivacyGroup ID for this collaboration",
            "metavar": "id",
        }
    )


@dataclass
class FBThreatExchangeCollabConfig(
    CollaborationConfigWithDefaults,
    _FBThreatExchangeCollabConfigRequiredFields,
):
    api: str = field(init=False, default=_API_NAME)
    # TODO - to restore someday in the future
    # app_token_override: t.Optional[str] = field(
    #     default=None,
    #     metadata={
    #         "help": "if you need to use a specific app for this collaboration",
    #         "metavar": "APP_TOKEN",
    #     },
    # )


@dataclass
class FBThreatExchangeCheckpoint(state.FetchCheckpointBase):
    """
    State about the progress of a /threat_updates-backed state.

    If a client does not resume tailing the threat_updates endpoint fast enough,
    deletion records will be removed, making it impossible to determine which
    records should be retained without refetching the entire dataset from scratch.
    """

    update_time: int = 0
    last_fetch_time: int = field(default_factory=lambda: int(time.time()))

    def is_stale(self) -> bool:
        """
        The API implementation will retain for 90 days

        https://developers.facebook.com/docs/threat-exchange/reference/apis/threat-updates/
        """
        return time.time() - self.last_fetch_time > 3600 * 24 * 85  # 85 days

    def get_progress_timestamp(self) -> int:
        return self.update_time


@dataclass
class FBThreatExchangeOpinion(state.SignalOpinion):

    REACTION_DESCRIPTOR_ID: t.ClassVar[int] = -1

    owner_app_id: int
    descriptor_id: t.Optional[int]

    def __setstate__(self, d: t.Dict[str, t.Any]) -> None:
        """Implemented for pickle version compatibility."""
        # 0.99.0 => 1.0.0:
        ### field 'owner_id' renamed to 'owner_app_id' and 'is_mine'
        if "owner" in d:
            d["owner_app_id"] = d["owner"]
        super().__setstate__(d)


@dataclass
class FBThreatExchangeIndicatorRecord(state.FetchedSignalMetadata):

    opinions: t.List[FBThreatExchangeOpinion]

    def get_as_opinions(
        self,
    ) -> t.List[FBThreatExchangeOpinion]:
        return self.opinions

    @classmethod
    def from_threatexchange_json(
        cls, my_app_id: int, te_json: ThreatUpdateJSON
    ) -> t.Optional["FBThreatExchangeIndicatorRecord"]:
        if te_json.should_delete:
            return None

        explicit_opinions: t.Dict[int, FBThreatExchangeOpinion] = {}
        implicit_opinions: t.Dict[int, state.SignalOpinionCategory] = {}

        for td_json in te_json.raw_json["descriptors"]["data"]:
            td_id = int(td_json["id"])
            if "owner" not in td_json:
                # Regrettable bug in threatexchange where descriptor from
                # `descriptors` is visible but `owner` field is not.
                # It's on Meta to fix this bug, but for now just drop the record.
                # This bug breaks the correctness of threat_updates, because
                # it can oscillate between visible and not separately from the
                # threat_update order.
                # TODO: Should we log this?
                continue
            owner_id = int(td_json["owner"]["id"])
            status = td_json["status"]
            # added_on = td_json["added_on"]
            tags = td_json.get("tags", [])
            # This is needed because ThreatExchangeAPI.get_threat_descriptors()
            # does a transform, but other locations do not
            if isinstance(tags, dict):
                tags = sorted(tag["text"] for tag in tags["data"])

            category = state.SignalOpinionCategory.INVESTIGATION_SEED

            if status == "MALICIOUS":
                category = state.SignalOpinionCategory.POSITIVE_CLASS
            elif status == "NON_MALICIOUS":
                category = state.SignalOpinionCategory.NEGATIVE_CLASS

            explicit_opinions[owner_id] = FBThreatExchangeOpinion(
                owner_id == my_app_id,
                category,
                set(tags),
                owner_id,
                td_id,
            )

            # See: https://developers.facebook.com/docs/threat-exchange/reference/apis/reaction-type/
            for reaction in td_json.get("reactions", []):
                rxn = reaction["key"]
                for owner in reaction["value"].split(","):
                    owner_id = int(owner)
                    if rxn == "HELPFUL":
                        implicit_opinions[
                            owner_id
                        ] = state.SignalOpinionCategory.POSITIVE_CLASS
                    elif (
                        rxn == "DISAGREE_WITH_TAGS"
                        and owner_id not in implicit_opinions
                    ):
                        implicit_opinions[
                            owner_id
                        ] = state.SignalOpinionCategory.NEGATIVE_CLASS

        for owner_id, category in implicit_opinions.items():
            if owner_id in explicit_opinions:
                continue
            explicit_opinions[owner_id] = FBThreatExchangeOpinion(
                owner_id == my_app_id,
                category,
                set(),
                owner_id,
                FBThreatExchangeOpinion.REACTION_DESCRIPTOR_ID,
            )

        if not explicit_opinions:
            # Visibility bug of some kind on TE API :(
            return None
        return cls(list(explicit_opinions.values()))

    def merge(self, other: "FBThreatExchangeIndicatorRecord") -> None:
        """
        Combine another indicator record with this one.

        This is needed when there are multiple records in ThreatExchange
        of equivalent types - i.e.
          * URI
          * RAW_URI
          * UNCLICKABLE_URL

        Most of the time, platforms record the exact same record for each,
        but it's not guaranteed.
        """
        # We could try to dedupe identical opinions, but instead just take
        # them all
        self.opinions.extend(other.opinions)

    @staticmethod
    def te_threat_updates_fields() -> t.Tuple[str, ...]:
        """The input to the "field" selector for the API"""
        return (
            "indicator",
            "type",
            "last_updated",
            "should_delete",
            "descriptors{%s}"
            % ",".join(
                (
                    "id",
                    "reactions",
                    "owner{id}",
                    "tags",
                    "status",
                )
            ),
        )


@dataclass
class FBThreatExchangeCredentials(auth.CredentialHelper):
    ENV_VARIABLE: t.ClassVar[str] = "TX_ACCESS_TOKEN"
    FILE_NAME: t.ClassVar[str] = "~/.txtoken"

    api_token: str

    @classmethod
    def _from_str(cls, s: str) -> t.Optional["FBThreatExchangeCredentials"]:
        return cls(s.strip())

    def _are_valid(self) -> bool:
        return is_valid_app_token(self.api_token)


ThreatExchangeDelta = state.FetchDelta[
    t.Tuple[str, str],
    FBThreatExchangeIndicatorRecord,
    FBThreatExchangeCheckpoint,
]


class FBThreatExchangeSignalExchangeAPI(
    auth.SignalExchangeWithAuth[
        FBThreatExchangeCollabConfig, FBThreatExchangeCredentials
    ],
    SignalExchangeAPIWithSimpleUpdates[
        FBThreatExchangeCollabConfig,
        FBThreatExchangeCheckpoint,
        FBThreatExchangeIndicatorRecord,
    ],
):
    def __init__(
        self, client: ThreatExchangeAPI, collab: FBThreatExchangeCollabConfig
    ) -> None:
        self.client = client
        self.collab = collab

    @classmethod
    def for_collab(
        cls,
        collab: FBThreatExchangeCollabConfig,
        credentials: t.Optional[FBThreatExchangeCredentials] = None,
    ) -> "FBThreatExchangeSignalExchangeAPI":
        credentials = credentials or FBThreatExchangeCredentials.get(cls)
        client = ThreatExchangeAPI(credentials.api_token)
        return cls(client, collab)

    @classmethod
    def get_name(cls) -> str:
        return _API_NAME

    @staticmethod
    def get_checkpoint_cls() -> t.Type[FBThreatExchangeCheckpoint]:
        return FBThreatExchangeCheckpoint

    @staticmethod
    def get_record_cls() -> t.Type[FBThreatExchangeIndicatorRecord]:
        return FBThreatExchangeIndicatorRecord

    @staticmethod
    def get_config_cls() -> t.Type[FBThreatExchangeCollabConfig]:
        return FBThreatExchangeCollabConfig

    @staticmethod
    def get_credential_cls() -> t.Type[FBThreatExchangeCredentials]:
        return FBThreatExchangeCredentials

    def fetch_iter(
        self,
        supported_signal_types: t.Sequence[t.Type[SignalType]],
        # None if fetching for the first time,
        # otherwise the previous FetchDelta returned
        checkpoint: t.Optional[FBThreatExchangeCheckpoint],
    ) -> t.Iterator[ThreatExchangeDelta]:
        start_time = None if checkpoint is None else checkpoint.update_time
        cursor = self.client.get_threat_updates(
            self.collab.privacy_group,
            start_time=start_time,
            page_size=100,
            fields=ThreatUpdateJSON.te_threat_updates_fields(),
            decode_fn=ThreatUpdateJSON,
        )
        type_mapping = _make_indicator_type_mapping(supported_signal_types)

        highest_time = start_time or 0
        batch: t.List[ThreatUpdateJSON]
        for batch in cursor:
            assert batch, "empty update?"
            highest_time = max(highest_time, max(update.time for update in batch))

            updates = {}
            for u in batch:
                converted = FBThreatExchangeIndicatorRecord.from_threatexchange_json(
                    self.client.app_id, u
                )
                updates[u.threat_type, u.indicator] = converted

            yield ThreatExchangeDelta(
                updates,
                FBThreatExchangeCheckpoint(highest_time),
            )

    @classmethod
    def naive_convert_to_signal_type(
        cls,
        signal_types: t.Sequence[t.Type[SignalType]],
        collab: FBThreatExchangeCollabConfig,
        fetched: t.Mapping[
            t.Tuple[str, str], t.Optional[FBThreatExchangeIndicatorRecord]
        ],
    ) -> t.Dict[t.Type[SignalType], t.Dict[str, FBThreatExchangeIndicatorRecord]]:
        """
        Convert ThreatExchange Indicator records to SignalTypes.

        We override this method from the base in order to make the signal type
        mapping just once.

        ThreatExchange uses a helper mixin that SignalTypes can implement in order
        to instruct the API how to convert ThreatExchange's ThreatType into
        a SignalType. ThreatExchange supports multiple ThreatTypes for the same
        SignalType, and so it's possible there are duplicate records. It's even
        possible that the uploader isn't consistent with their labeling for the
        "identical" records in ThreatExchange.
        """
        ret: t.Dict[
            t.Type[SignalType], t.Dict[str, FBThreatExchangeIndicatorRecord]
        ] = {}
        mapping = _make_indicator_type_mapping(signal_types)

        for (type_str, signal_str), metadata in fetched.items():
            if metadata is None:
                continue
            potential_types = mapping.get(type_str)
            if potential_types is None:
                continue
            indicator_tags = {t for opinion in metadata.opinions for t in opinion.tags}
            for tag, s_types in potential_types.items():
                if tag is not None and tag not in indicator_tags:
                    continue
                for tx_s_type in s_types:
                    s_type_specific_signal_str = (
                        tx_s_type.normalize_fb_threatexchange_indicator(
                            type_str, signal_str, tag
                        )
                    )
                    s_type = t.cast(t.Type[SignalType], tx_s_type)
                    inner = ret.get(s_type)
                    if inner is None:
                        inner = {}
                        ret[s_type] = inner
                    to_insert = _merge_record_for_signal_type(
                        metadata, tag, inner.get(s_type_specific_signal_str)
                    )
                    if to_insert is not None:
                        inner[s_type_specific_signal_str] = to_insert

        return ret


def _merge_record_for_signal_type(
    tx_record: FBThreatExchangeIndicatorRecord,
    tag: t.Optional[str],
    existing: t.Optional[FBThreatExchangeIndicatorRecord],
) -> t.Optional[FBThreatExchangeIndicatorRecord]:
    if tag is not None:
        applicable_opinions = [
            o for o in tx_record.opinions if any(t in tag for t in o.tags)
        ]
        if not applicable_opinions:
            return None
        if len(applicable_opinions) != len(tx_record.opinions):
            tx_record = FBThreatExchangeIndicatorRecord(applicable_opinions)
    if existing is not None:
        existing.merge(tx_record)
        return None
    return tx_record


def _make_indicator_type_mapping(
    supported_signal_types: t.Sequence[t.Type[SignalType]],
) -> t.Mapping[
    str,
    t.Mapping[t.Optional[str], t.Sequence[t.Type[HasFbThreatExchangeIndicatorType]]],
]:
    """
    Based on the given signal types, create a map for converting ThreatIndicators.

    The returned mapping is ThreatType => ?tag => SignalType.

    For example, with MD5, and one test type:
    ```
    {
       "HASH_VIDEO_MD5": {
           None: [VideoMd5Signal],
        },
        "HASH_MD5": {
           "media_type_video": [VideoMD5Signal]
        }
        "DEBUG_STRING": {
            "type:foo": [FooType],
        }
    }
    ```
    """
    ret: t.DefaultDict[
        str,
        t.DefaultDict[
            t.Optional[str], t.List[t.Type[HasFbThreatExchangeIndicatorType]]
        ],
    ] = defaultdict(lambda: defaultdict(list))
    for st in supported_signal_types:
        if not issubclass(st, HasFbThreatExchangeIndicatorType):
            continue
        types = st.INDICATOR_TYPE
        if isinstance(types, str):
            types = {types: None}
        elif isinstance(types, set):
            types = {tag: None for tag in types}
        else:
            assert isinstance(types, dict)
        for type_, tag in types.items():
            ret[type_][tag].append(st)

    return ret



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import json
import pathlib
import dacite
import dataclasses
from enum import Enum
import typing as t

T = t.TypeVar("T")


def dataclass_dump_file(path: pathlib.Path, obj) -> None:
    with path.open("w") as fp:
        dataclass_dump(fp, obj)


def _as_dict(obj: t.Any) -> t.Dict[str, t.Any]:
    json_dict = dataclasses.asdict(obj)
    # Sanity check - we want to make sure it will also come out the other end
    # And this will wrong-type error if it can't
    obj_sanity_check = dataclass_load_dict(json_dict, obj.__class__)
    assert obj == obj_sanity_check, "object changed during serialization?"
    return json_dict


def dataclass_dump(fp: t.IO[str], obj) -> None:
    json_dict = _as_dict(obj)
    return json.dump(json_dict, fp, indent=2, default=_json_cast_default)


def dataclass_dumps(obj) -> str:
    json_dict = _as_dict(obj)
    return json.dumps(json_dict, indent=2, default=_json_cast_default)


def dataclass_load_file(
    path: pathlib.Path, cls: t.Type[T], *, default: t.Optional[T] = None
) -> T:
    if not path.is_file():
        if default is not None:
            return default
        raise ValueError(f"cannot load dataclass: no such file {path}")
    with path.open("r") as fp:
        return dataclass_load(fp, cls)


def dataclass_load(fp: t.IO[str], cls: t.Type[T]) -> T:
    json_dict = json.load(fp)
    return dataclass_load_dict(json_dict, cls)


def dataclass_loads(s: str, cls: t.Type[T]) -> T:
    json_dict = json.loads(s)
    return dataclass_load_dict(json_dict, cls)


def dataclass_load_dict(json_dict: t.Dict[str, t.Any], cls: t.Type[T]) -> T:
    return dacite.from_dict(
        data_class=cls,
        data=json_dict,
        config=dacite.Config(cast=[Enum, set, tuple]),
    )


def _json_cast_default(obj):
    if isinstance(obj, set):
        return list(obj)
    if isinstance(obj, Enum):
        return obj.value
    raise TypeError



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
The SignalExchangeAPI talks to external APIs to read/write signals

@see SignalExchangeAPI
"""

from abc import ABC, abstractmethod
import typing as t

from threatexchange import common
from threatexchange.exchanges.collab_config import CollaborationConfigBase
from threatexchange.signal_type.signal_base import SignalType
from threatexchange.exchanges import fetch_state as state

TCollabConfig = t.TypeVar("TCollabConfig", bound=CollaborationConfigBase)


class SignalExchangeAPI(
    t.Generic[
        TCollabConfig,
        state.TFetchCheckpoint,
        state.TFetchedSignalMetadata,
        state.TUpdateRecordKey,
        state.TUpdateRecordValue,
    ],
    ABC,
):
    """
    APIs to read and maybe write signals.

    SignalExchangeAPIs should checkpoint their progress, so that they
    can tail updates. If this is not possible, they can instead use
    checkpoints to record how long it has been since a full fetch, and
    trigger a refresh if a certain amount of time has passed.

    While this interface is primarily intended for connecting with
    externally hosted servers, it might be useful to write adopters for
    certain formats of local files, which could be valuable for testing.

    Is assumed that fetched signals have some metadata attached to them,
    which is unique to that API. Additionally, it a assumed that there
    might be multiple contributors (owners) to signals inside of an API.

    An instance of this class can retain state (caching connecting, etc)
    as needed, and objects may be persisted to fetch multiple configs.

    = On fetch_iter() returns =
    In order to efficiently store state, it's assumed that data pulled from
    the API can be partitioned in some way by key. If this doesn't make sense
    for your API, or it's mostly a toy implementation, empty string is a
    valid key, and the value can be the entire dataset.

    = On Authentification =
    It's expected that an instance of the class is fully authenticated,
    and in general, auth should be passible by the constructor.

    If an API is constructed via the for_collab classmethod constructor,
    it should attempt to authenticate itself via discovering it from the
    execution environment and passing it via __init__(). If constructed
    directly, it should not search for credentials.

    """

    @classmethod
    @abstractmethod
    def for_collab(cls, collab: TCollabConfig) -> "SignalExchangeAPI":
        """
        An alternative constructor to get a working instance of the API.

        An instance provided by this class is expected to:
        1. Be fully authenticated or throw an exception
        2. All methods are callable if supported by the base implementation

        For implementations that do need additional configuration or
        authentification, @see threatexchange.exchanges.auth
        If not explicitly given auth by those helpers, the implementation
        should attempt to find its own, and throw an exception otherwise.

        Don't:
        * Prompt the user via stdin
        * Return an instance that will throw when any methods are called

        Usages of this library will only attempt to instanciate an API
        when they have an API call to make - all methods for manipulating
        local state are classmethods.
        """
        raise NotImplementedError

    @classmethod
    def get_name(cls) -> str:
        """
        A simple string name unique to SignalExchangeAPIs in use.

        It should be one lowercase_with_underscores word.

        This shouldn't be changed once comitted, or you may break naive
        storage solutions (like the one in the CLI) which stores fetched
        data by (SignalExchangeAPI.name(), collab_name).
        """
        name = cls.__name__
        for suffix in ("API", "Exchange"):
            if name.endswith(suffix):
                name = name[: -len(suffix)]
        return common.class_name_to_human_name(name, "Signal")

    @staticmethod
    @abstractmethod
    def get_checkpoint_cls() -> t.Type[state.TFetchCheckpoint]:
        """Returns the dataclass used to control checkpoint for this API"""
        pass

    @staticmethod
    @abstractmethod
    def get_record_cls() -> t.Type[state.TFetchedSignalMetadata]:
        """Returns the dataclass used to store records for this API"""
        pass

    @staticmethod
    @abstractmethod
    def get_config_cls() -> t.Type[TCollabConfig]:
        """Returns the dataclass used to store records for this API"""
        pass

    @classmethod
    def fetch_value_merge(
        cls,
        old: t.Optional[state.TUpdateRecordValue],
        new: t.Optional[state.TUpdateRecordValue],  # can be modified in-place
    ) -> t.Optional[state.TUpdateRecordValue]:
        """
        Merge a new update produced by fetch.

        Returning a value of None indicates that the entry should be deleted.

        Most implementations will probably prefer the default, which is to
        replace the record entirely.

        It is safe to mutate `new` inline and return it, if needed.
        """
        # Default implementation is replace
        return new

    @classmethod
    @t.final
    def naive_fetch_merge(
        cls,
        old: t.Dict[
            state.TUpdateRecordKey, state.TUpdateRecordValue
        ],  # modified in place
        new: t.Mapping[state.TUpdateRecordKey, t.Optional[state.TUpdateRecordValue]],
    ) -> None:
        """
        Merge the results of a fetch in-memory.

        This implementation is mostly for demonstration purposes and testing,
        since even simple usecases may prefer to avoid loading the whole dataset
        in memory and merging by key.

        For example, if you have nothing else, merging NCMEC update records
        together keyed by ID will eventually get you an entire copy of the database.
        """
        for k, v in new.items():
            new_v = cls.fetch_value_merge(old.get(k), v)
            if new_v is None:
                old.pop(k, None)  # type: ignore
            else:
                old[k] = new_v

    @classmethod
    @abstractmethod
    def naive_convert_to_signal_type(
        cls,
        signal_types: t.Sequence[t.Type[SignalType]],
        collab: TCollabConfig,
        fetched: t.Mapping[state.TUpdateRecordKey, state.TUpdateRecordValue],
    ) -> t.Dict[t.Type[SignalType], t.Dict[str, state.TFetchedSignalMetadata]]:
        """
        Convert the record from the API format to the format needed for indexing.

        This is the fallback method of creating state when there isn't a
        specialized storage for the fetch type.

        """
        raise NotImplementedError

    @abstractmethod
    def fetch_iter(
        self,
        supported_signal_types: t.Sequence[t.Type[SignalType]],
        # None if fetching for the first time,
        # otherwise the previous FetchDelta returned
        checkpoint: t.Optional[state.TFetchCheckpoint],
    ) -> t.Iterator[
        state.FetchDelta[
            state.TUpdateRecordKey, state.TUpdateRecordValue, state.TFetchCheckpoint
        ]
    ]:
        """
        Call out to external resources, fetching a batch of updates per yield.

        Many APIs are a sequence of events: (creates/updates, deletions)
        In that case, it's important the these events are strictly ordered.
        I.e. if the sequence is create => delete, if the sequence is reversed
        to delete => create, the end result is a stored record, when the
        expected is a deleted one.

        Updates are assumed to have a keys that can partition the dataset. See the
        note on this in the class docstring.

        The iterator may be abandoned before it is completely exhausted.

        If the iterator returns, it should be because there is no more data
        (i.e. the fetch is up to date).
        """
        raise NotImplementedError

    # TODO - Restore in a future version
    # def report_seen(
    #     self,
    #     collab: TCollabConfig,
    #     s_type: SignalType,
    #     signal: str,
    #     metadata: state.TFetchedSignalMetadata,
    # ) -> None:
    #     """
    #     Report that you observed this signal.

    #     This is an optional API, and places that use it should catch
    #     the NotImplementError.
    #     """
    #     raise NotImplementedError

    def report_opinion(
        self,
        s_type: t.Type[SignalType],
        signal: str,
        opinion: state.SignalOpinion,
    ) -> None:
        """
        Weigh in on a signal for this collaboration.

        Most implementations will want a full replacement specialization, but this
        allows a common interface for all uploads for the simplest usecases.

        This is an optional API, and places that use it should catch
        the NotImplementError.
        """
        raise NotImplementedError


# Convenience for avoiding mypy strict errors
AnySignalExchangeAPI = SignalExchangeAPI[t.Any, t.Any, t.Any, t.Any, t.Any]

# A convenience helper since mypy can't intuit that bound != t.Any
# For methods like get_checkpoint_cls
TSignalExchangeAPI = SignalExchangeAPI[
    CollaborationConfigBase,
    state.FetchCheckpointBase,
    state.FetchedSignalMetadata,
    t.Any,
    t.Any,
]

TSignalExchangeAPICls = t.Type[TSignalExchangeAPI]


class SignalExchangeAPIWithSimpleUpdates(
    SignalExchangeAPI[
        TCollabConfig,
        state.TFetchCheckpoint,
        state.TFetchedSignalMetadata,
        t.Tuple[str, str],
        state.TFetchedSignalMetadata,
    ]
):
    """
    An API that conveniently maps directly into the form needed by index.

    If the API supports returning exactly the hashes and all the metadata needed
    to make a decision on the hash without needing an indirection of ID (for example,
    to support deletes), then you can choose to directly return it in a form that
    maps directly into SignalType.
    """

    @classmethod
    def naive_convert_to_signal_type(
        cls,
        signal_types: t.Sequence[t.Type[SignalType]],
        collab: TCollabConfig,
        fetched: t.Mapping[t.Tuple[str, str], t.Optional[state.TFetchedSignalMetadata]],
    ) -> t.Dict[t.Type[SignalType], t.Dict[str, state.TFetchedSignalMetadata]]:
        ret: t.Dict[t.Type[SignalType], t.Dict[str, state.TFetchedSignalMetadata]] = {}
        type_by_name = {st.get_name(): st for st in signal_types}
        for (type_str, signal_str), metadata in fetched.items():
            s_type = type_by_name.get(type_str)
            if s_type is None or metadata is None:
                continue
            inner = ret.get(s_type)
            if inner is None:
                inner = {}
                ret[s_type] = inner
            inner[signal_str] = metadata
        return ret



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the URL MD5 signal types.
"""

import typing as t
import hashlib
from threatexchange.content_type.content_base import ContentType
from threatexchange.content_type.url import URLContent

from threatexchange.signal_type import signal_base
from threatexchange import common
from threatexchange.signal_type.url import URLSignal
from threatexchange.exchanges.impl.fb_threatexchange_signal import (
    HasFbThreatExchangeIndicatorType,
)


class UrlMD5Signal(
    signal_base.SimpleSignalType,
    signal_base.TextHasher,
    HasFbThreatExchangeIndicatorType,
):
    """
    Simple signal type for URL MD5s.
    """

    INDICATOR_TYPE = "HASH_URL_MD5"

    @classmethod
    def get_content_types(cls) -> t.List[t.Type[ContentType]]:
        return [URLContent]

    @classmethod
    def get_index_cls(cls) -> t.Type[signal_base.TrivialSignalTypeIndex]:
        return signal_base.TrivialSignalTypeIndex

    @classmethod
    def hash_from_str(cls, url: str) -> str:
        encoded_url = common.normalize_url(url)
        url_hash = hashlib.md5(encoded_url)
        return url_hash.hexdigest()

    @staticmethod
    def get_examples() -> t.List[str]:
        return [UrlMD5Signal.hash_from_str(s) for s in URLSignal.get_examples()]



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

BITS_IN_PDQ = 256
PDQ_HEX_STR_LEN = int(BITS_IN_PDQ / 4)


def simple_distance_binary(bin_a: str, bin_b: str) -> int:
    """
    Returns the hamming distance of two binary strings.
    """
    assert len(bin_a) == BITS_IN_PDQ
    assert len(bin_b) == BITS_IN_PDQ
    return sum(bin_a[i] != bin_b[i] for i in range(BITS_IN_PDQ))


def simple_distance(hex_a: str, hex_b: str) -> int:
    """
    Returns the binary hamming distance of two hexadecimal strings.
    """
    return simple_distance_binary(hex_to_binary_str(hex_a), hex_to_binary_str(hex_b))


def hex_to_binary_str(pdq_hex: str) -> str:
    """
    Convert a hexadecimal string to a binary string. Requires input string to be length PDQ_HEX_STR_LEN.
    """
    assert len(pdq_hex) == PDQ_HEX_STR_LEN
    # padding to 4 bindigits each hexdigit
    result = "".join(bin(int(c, 16))[2:].zfill(4) for c in pdq_hex)
    assert len(result) == BITS_IN_PDQ
    return result


def binary_str_to_hex(pdq_binary: str) -> str:
    """
    Convert a binary string to a hexadecimal string. Requires input string to be length BITS_IN_PDQ.
    """
    assert len(pdq_binary) == BITS_IN_PDQ
    # [2:] ignores the 0x at the begining of the hex_str
    result = hex(int(pdq_binary, 2))[2:].zfill(PDQ_HEX_STR_LEN)
    assert len(result) == PDQ_HEX_STR_LEN
    return result


def pdq_match(pdq_hex_a: str, pdq_hex_b: str, threshold: int) -> bool:
    """
    Returns true if hamming distance of two hex strings are within (<=) the given threshold.
    """
    distance = simple_distance(pdq_hex_a, pdq_hex_b)
    return distance <= threshold



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import pytest
import pickle
import random
from threatexchange.extensions.vpdq.vpdq_util import VpdqCompactFeature

from threatexchange.signal_type.index import IndexMatch

try:
    import vpdq

    _DISABLED = False
except (ImportError, ModuleNotFoundError) as e:
    _DISABLED = True
else:
    import typing as t
    from threatexchange.extensions.vpdq.vpdq_index import VPDQIndex, VPDQSimilarityInfo
    from threatexchange.extensions.vpdq.vpdq_util import (
        json_to_vpdq,
        prepare_vpdq_feature,
        VPDQ_QUALITY_THRESHOLD,
        vpdq_to_json,
        dedupe,
        quality_filter,
    )
    from threatexchange.extensions.vpdq.tests.utils import (
        get_random_vpdq_features,
        pdq_hashes_to_vpdq_features,
    )
    from threatexchange.extensions.vpdq.vpdq import VPDQSignal
    from threatexchange.tests.hashing.utils import (
        get_random_hash,
        get_similar_hash,
        get_zero_hash,
    )

pytestmark = pytest.mark.skipif(_DISABLED, reason="vpdq not installed")


def pdq_to_vpdq(hashes):
    return vpdq_to_json(
        [VpdqCompactFeature(h, 100, float(ts)) for ts, h in enumerate(hashes)]
    )


EXAMPLE_META_DATA = {"hash_type": "vpdq", "video_id": 5}
VIDEO1_META_DATA = object()
VIDEO2_META_DATA = object()
VIDEO3_META_DATA = object()
VIDEO4_META_DATA = object()
# Frames/features
F1 = get_similar_hash(get_zero_hash(), 16)
F2 = get_similar_hash(get_zero_hash(), 16 + (32 + 1))
F3 = get_similar_hash(get_zero_hash(), 16 + (32 + 1) * 2)
F4 = get_similar_hash(get_zero_hash(), 16 + (32 + 1) * 3)
# Groups
G1 = [get_similar_hash(F1, i) for i in range(0, 16)]
G2 = [get_similar_hash(F2, i) for i in range(0, 16)]
G3 = [get_similar_hash(F3, i) for i in range(0, 16)]

HASH = pdq_to_vpdq([F1, F2, F3, F4])
FEATURES = prepare_vpdq_feature(HASH, VPDQ_QUALITY_THRESHOLD)
# Three groups of hashes that are 128 hamming distance away from each other.
# The hashes match with each other within the group. And each group's hashes don't match.


def test_utils():
    example_hash = VPDQSignal.get_examples()[0]
    example_features = json_to_vpdq(example_hash)
    assert example_hash == vpdq_to_json(example_features)

    feature_size = len(example_features)
    assert len(dedupe(example_features)) == feature_size
    assert len(quality_filter(example_features, 50)) == feature_size
    assert len(quality_filter(example_features, 101)) == 0
    example_features.append(example_features[0])
    assert len(dedupe(example_features)) == feature_size

    frame_counts = 100
    features = get_random_vpdq_features(frame_counts)
    assert len(features) == frame_counts
    VPDQSignal.validate_signal_str(vpdq_to_json(features))
    pdq_hashes = [get_random_hash() for i in range(frame_counts)]
    VPDQSignal.validate_signal_str(
        vpdq_to_json(pdq_hashes_to_vpdq_features(pdq_hashes))
    )


def test_simple():
    index = VPDQIndex.build([[HASH, EXAMPLE_META_DATA]])
    assert index._entry_idx_to_features_and_entries[0][0] == FEATURES
    assert len(index._index_idx_to_vpdqHex_and_entry) == len(FEATURES)
    res = index.query(HASH)
    # A complete match to itself
    assert res[0] == IndexMatch(VPDQSimilarityInfo(100.0, 100.0), EXAMPLE_META_DATA)


def test_half_match():
    index = VPDQIndex.build([[HASH, EXAMPLE_META_DATA]], query_match_threshold_pct=0)
    half_hash = FEATURES[0 : int(len(FEATURES) / 2)]
    res = index.query(vpdq_to_json(half_hash))
    assert res[0] == IndexMatch(VPDQSimilarityInfo(100.0, 50.0), EXAMPLE_META_DATA)

    index = VPDQIndex.build(
        [[vpdq_to_json(half_hash), EXAMPLE_META_DATA]], query_match_threshold_pct=0
    )
    res = index.query(HASH)
    assert res[0] == IndexMatch(VPDQSimilarityInfo(50.0, 100.0), EXAMPLE_META_DATA)


def test_serialize():
    index = VPDQIndex.build([[HASH, EXAMPLE_META_DATA]])
    pickled_data = pickle.dumps(index)
    reconstructed_index = pickle.loads(pickled_data)
    res = reconstructed_index.query(HASH)
    assert res[0] == IndexMatch(VPDQSimilarityInfo(100.0, 100.0), EXAMPLE_META_DATA)


def test_empty_video_():
    empty_video = []
    video1 = pdq_hashes_to_vpdq_features(G1)

    index = VPDQIndex.build([[vpdq_to_json(video1), VIDEO1_META_DATA]])
    res = index.query(vpdq_to_json(empty_video))
    assert len(res) == 0
    with pytest.raises(ValueError):
        VPDQIndex.build([[vpdq_to_json(empty_video), VIDEO2_META_DATA]])


def test_empty_signal_str():
    empty_signal_str = ""
    video1 = pdq_hashes_to_vpdq_features(G1)
    index = VPDQIndex.build([[vpdq_to_json(video1), VIDEO1_META_DATA]])
    res = index.query(empty_signal_str)
    assert len(res) == 0
    with pytest.raises(ValueError):
        VPDQIndex.build([[vpdq_to_json(empty_signal_str), VIDEO2_META_DATA]])


def test_duplicate_hashes():
    index = VPDQIndex.build([[HASH, VIDEO1_META_DATA]])
    index.add(HASH, VIDEO2_META_DATA)
    res = index.query(HASH)
    # A complete match to itself
    assert res[0] == IndexMatch(VPDQSimilarityInfo(100.0, 100.0), VIDEO1_META_DATA)
    assert res[1] == IndexMatch(VPDQSimilarityInfo(100.0, 100.0), VIDEO2_META_DATA)


def test_no_match():
    video1 = pdq_hashes_to_vpdq_features(G1)
    video2 = pdq_hashes_to_vpdq_features(G2)

    index = VPDQIndex.build([[vpdq_to_json(video1), VIDEO1_META_DATA]])
    res = index.query(vpdq_to_json(video2))
    assert len(res) == 0

    index = VPDQIndex.build([[vpdq_to_json(video2), VIDEO2_META_DATA]])
    res = index.query(vpdq_to_json(video1))
    assert len(res) == 0


def test_no_match_with_zero_threshold():
    # Index should not return any non-match result even if threshold is zero
    video1 = pdq_hashes_to_vpdq_features(G1)
    video2 = pdq_hashes_to_vpdq_features(G2)
    index = VPDQIndex.build(
        [[vpdq_to_json(video1), VIDEO1_META_DATA]],
        query_match_threshold_pct=0,
        index_match_threshold_pct=0,
    )
    res = index.query(vpdq_to_json(video2))
    assert len(res) == 0


def test_match_below_and_above_80pct_query_threshold():
    # Video1 contains five out of ten frames that match with indexed video4 -> query percent 50% and filtered
    # Video2 contains eight out of ten frames that match with indexed video4 -> query percent 80% and return
    # Video3 contains nine out of ten frames that match with indexed video4 -> query percent 90% and return
    video1 = pdq_hashes_to_vpdq_features(
        random.sample(G1, 2) + random.sample(G2, 3) + random.sample(G3, 5)
    )
    video2 = pdq_hashes_to_vpdq_features(
        random.sample(G1, 1) + random.sample(G2, 1) + random.sample(G3, 8)
    )
    video3 = pdq_hashes_to_vpdq_features(random.sample(G1, 1) + random.sample(G3, 9))
    video4 = pdq_hashes_to_vpdq_features(random.sample(G3, 10))

    index = VPDQIndex.build(
        [[vpdq_to_json(video4), VIDEO4_META_DATA]],
        query_match_threshold_pct=80,
    )
    res = index.query(vpdq_to_json(video1))
    assert len(res) == 0

    res = index.query(vpdq_to_json(video2))
    assert res[0] == IndexMatch(VPDQSimilarityInfo(80.0, 100.0), VIDEO4_META_DATA)

    res = index.query(vpdq_to_json(video3))
    assert res[0] == IndexMatch(VPDQSimilarityInfo(90.0, 100.0), VIDEO4_META_DATA)


def test_match_below_and_above_80pct_index_threshold():
    # Indexed video1 contains five out of ten frames that match with quered video4 -> index percent 50% and filtered
    # Indexed video2 contains eight out of ten frames that match with quered video4 -> index percent 80% and return
    # Indexed video3 contains nine out of ten frames that match with quered video4 -> index percent 90% and return

    video1 = pdq_hashes_to_vpdq_features(
        random.sample(G1, 2) + random.sample(G2, 3) + random.sample(G3, 5)
    )
    video2 = pdq_hashes_to_vpdq_features(
        random.sample(G1, 1) + random.sample(G2, 1) + random.sample(G3, 8)
    )
    video3 = pdq_hashes_to_vpdq_features(random.sample(G1, 1) + random.sample(G3, 9))
    video4 = pdq_hashes_to_vpdq_features(random.sample(G3, 10))

    index = VPDQIndex.build(
        [[vpdq_to_json(video1), VIDEO1_META_DATA]],
        index_match_threshold_pct=80,
    )
    res = index.query(vpdq_to_json(video4))
    assert len(res) == 0

    index = VPDQIndex.build(
        [[vpdq_to_json(video2), VIDEO2_META_DATA]],
        index_match_threshold_pct=80,
    )
    res = index.query(vpdq_to_json(video4))
    assert res[0] == IndexMatch(VPDQSimilarityInfo(100.0, 80.0), VIDEO2_META_DATA)

    index = VPDQIndex.build(
        [[vpdq_to_json(video3), VIDEO3_META_DATA]],
        index_match_threshold_pct=80,
    )
    res = index.query(vpdq_to_json(video4))
    assert res[0] == IndexMatch(VPDQSimilarityInfo(100.0, 90.0), VIDEO3_META_DATA)


def test_matches():
    # Two videos (length 10) each with one unmatched frame. Query video2 with index built from video1 and vice versa.
    # Delete the unmatched frame from video1 and query. Then Delete the unmatched frame from video2 and query.
    video1 = pdq_hashes_to_vpdq_features(random.sample(G1, 9))
    video2 = pdq_hashes_to_vpdq_features(random.sample(G1, 9))
    video1.append(VpdqCompactFeature(get_similar_hash(get_zero_hash(), 255), 100, 10.0))
    video2.append(
        VpdqCompactFeature(get_similar_hash(video1[0].pdq_hex, 128), 100, 10.0)
    )
    index = VPDQIndex.build([[vpdq_to_json(video1), VIDEO1_META_DATA]])
    res = index.query(vpdq_to_json(video2))
    assert res[0] == IndexMatch(VPDQSimilarityInfo(90.0, 90.0), VIDEO1_META_DATA)

    index = VPDQIndex.build([[vpdq_to_json(video2), VIDEO2_META_DATA]])
    res = index.query(vpdq_to_json(video1))
    assert res[0] == IndexMatch(VPDQSimilarityInfo(90.0, 90.0), VIDEO2_META_DATA)

    del video1[-1]
    index = VPDQIndex.build([[vpdq_to_json(video1), VIDEO1_META_DATA]])
    res = index.query(vpdq_to_json(video2))
    assert res[0] == IndexMatch(VPDQSimilarityInfo(90.0, 100.0), VIDEO1_META_DATA)

    del video2[-1]
    index = VPDQIndex.build([[vpdq_to_json(video1), VIDEO1_META_DATA]])
    res = index.query(vpdq_to_json(video2))

    assert res[0] == IndexMatch(VPDQSimilarityInfo(100.0, 100.0), VIDEO1_META_DATA)


def test_duplicate_matches():
    # There are video1 (length 10) with one duplicate frame and Video2 (length 9) where matches all
    # unique frames in video1.The match percentage is (100, 100) because the duplicate one is deduped and not counted.
    video1 = pdq_hashes_to_vpdq_features(random.sample(G1, 9))
    video2 = pdq_hashes_to_vpdq_features(
        [get_similar_hash(h.pdq_hex, 31) for h in video1]
    )
    video1.append(VpdqCompactFeature(video1[8].pdq_hex, 10, 9.0))

    index = VPDQIndex.build([[vpdq_to_json(video1), VIDEO1_META_DATA]])
    res = index.query(vpdq_to_json(video2))
    assert res[0] == IndexMatch(VPDQSimilarityInfo(100.0, 100.0), VIDEO1_META_DATA)


def test_duplicate_video_matches():
    # There are video1 with 10 frames (from G1 and G2 evenly) and video2 which contains duplicated
    # first five frames in video2. Video3's frames are matched with video1's that belong to same group.
    video1 = pdq_hashes_to_vpdq_features(random.sample(G1, 5) + random.sample(G2, 5))
    video2 = video1[0:5]
    video3 = pdq_hashes_to_vpdq_features(random.sample(G1, 5) + random.sample(G2, 5))

    index = VPDQIndex(query_match_threshold_pct=0)
    index.add_all(
        [
            [vpdq_to_json(video1), VIDEO1_META_DATA],
            [vpdq_to_json(video2), VIDEO2_META_DATA],
        ]
    )
    res = index.query(vpdq_to_json(video3))
    assert res[0] == IndexMatch(VPDQSimilarityInfo(100.0, 100.0), VIDEO1_META_DATA)
    assert res[1] == IndexMatch(VPDQSimilarityInfo(50.0, 100.0), VIDEO2_META_DATA)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import collections
import logging
import datetime
import logging
import time
import typing as t

from threatexchange.cli.cli_config import CLISettings
from threatexchange.cli.dataset_cmd import DatasetCommand
from threatexchange.exchanges.collab_config import CollaborationConfigBase
from threatexchange.exchanges.signal_exchange_api import (
    SignalExchangeAPI,
    TSignalExchangeAPICls,
)
from threatexchange.exchanges.fetch_state import (
    FetchCheckpointBase,
    FetchDeltaTyped,
    FetchedStateStoreBase,
)
from threatexchange.cli import command_base


class FetchCommand(command_base.Command):
    """
    Download content from signal exchange APIs to disk.

    Fetches data for collaborations that you've configured with
      $ threatexchange config collab edit

    Fetching is checkpointed, and you can safely interrupt by using CTRL+C.

    Some APIs require that you regularly re-fetch them within some number
    of days or else the stored data becomes invalid.

    This command is laconic by default, try `threatexchange -v fetch` for
    more granular details.
    """

    PROGRESS_PRINT_INTERVAL_SEC = 30

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap) -> None:
        ap.add_argument(
            "--clear",
            action="store_true",
            help="delete fetched state and checkpoints "
            "(you almost never need to do this)",
        )
        ap.add_argument(
            "--skip-index-rebuild",
            action="store_true",
            help="don't rebuild indices after fetch",
        )
        ap.add_argument("--limit", type=int, help="stop after fetching this many items")
        ap.add_argument(
            "--time-limit-sec",
            type=int,
            metavar="SEC",
            help="stop fetching after this many seconds",
        )
        ap.add_argument(
            "--checkpoint-every",
            type=int,
            metavar="SEC",
            default=600,
            help="write state to disk if still fetching, or 0 to disable",
        )
        ap.add_argument(
            "--only-api",
            choices=[api.get_name() for api in settings.apis],
            help="only fetch from this API",
        )
        ap.add_argument(
            "--only-collab",
            metavar="NAME",
            help="only fetch for this collaboration",
        )

    def __init__(
        self,
        # Defaults to make it easier to call from match
        clear: bool = False,
        time_limit_sec: t.Optional[int] = None,
        limit: t.Optional[int] = None,
        skip_index_rebuild: bool = False,
        only_api: t.Optional[str] = None,
        only_collab: t.Optional[str] = None,
        checkpoint_every: int = 600,
    ) -> None:
        self.clear = clear
        self.time_limit_sec = time_limit_sec
        self.limit = limit
        self.skip_index_rebuild = skip_index_rebuild
        self.only_api = only_api
        self.only_collab = only_collab
        self.collabs: t.List[CollaborationConfigBase] = []

        # Limits
        self.total_fetched_count = 0
        self.start_time = time.time()
        self.last_checkpoint_time = time.time()
        self.checkpoint_every = checkpoint_every

        # Progress
        self.last_update_time: t.Optional[int] = None
        # Print first update after 5 seconds
        self.last_update_printed = time.time() - self.PROGRESS_PRINT_INTERVAL_SEC + 5
        self.progress_fetched_count = 0
        self.counts: t.Dict[str, int] = collections.Counter()

    def has_hit_limits(self):
        if self.limit is not None and self.total_fetched_count >= self.limit:
            return True
        if self.time_limit_sec is not None:
            if time.time() - self.start_time >= self.time_limit_sec:
                return True
        return False

    def should_checkpoint(self):
        if self.checkpoint_every <= 0:
            return False
        return time.time() > self.checkpoint_every + self.last_checkpoint_time

    def execute(self, settings: CLISettings) -> None:
        # Verify collab arguments
        self.collabs = settings.get_all_collabs()
        if self.only_collab:
            self.collabs = [c for c in self.collabs if c.name == self.only_collab]
            if not self.collabs:
                raise command_base.CommandError(
                    f"No such collab '{self.only_collab}'", 2
                )
        if all(not c.enabled for c in self.collabs):
            self.stderr("All collabs are disabled. Nothing to do.")
            return

        # Do work
        if self.clear:
            self.stderr("Clearing fetched state")
            for api in settings.apis:
                store = settings.fetched_state.get_for_api(api)
                for collab in self.collabs:
                    if self.only_collab not in (None, collab.name):
                        continue
                    logging.info("Clearing %s - %s", api.get_name(), collab.name)
                    store.clear(collab)
            return

        all_succeeded = True
        any_succeded = False

        for api in settings.apis:
            logging.info("Fetching all %s's configs", api.get_name())
            succeeded = self.execute_for_api(settings, api)
            all_succeeded &= succeeded
            any_succeded |= succeeded

        if any_succeded and not self.skip_index_rebuild:
            self.stderr("Rebuilding match indices...")
            DatasetCommand().execute_generate_indices(settings)

        if not all_succeeded:
            raise command_base.CommandError("Some collabs had errors!", 3)

    def execute_for_api(
        self, settings: CLISettings, api: TSignalExchangeAPICls
    ) -> bool:
        success = True
        for collab in self.collabs:
            if collab.api != api.get_name():
                continue
            if not collab.enabled:
                logging.debug(
                    "Skipping %s, disabled",
                )
                continue
            fetch_ok = self.execute_for_collab(settings, collab)
            success &= fetch_ok
        return success

    def execute_for_collab(
        self,
        settings: CLISettings,
        collab: CollaborationConfigBase,
    ) -> bool:
        api = settings.apis.get_instance_for_collab(collab)
        store = settings.fetched_state.get_for_api(api.__class__)
        checkpoint = self._verify_store_and_checkpoint(store, collab)

        self.progress_fetched_count = 0
        self.current_collab = collab.name
        self.current_api = api.get_name()
        completed = False

        try:
            it = api.fetch_iter(settings.get_all_signal_types(), checkpoint)
            delta: FetchDeltaTyped
            for delta in it:
                assert delta.checkpoint is not None  # Infinite loop protection
                progress_time = delta.checkpoint.get_progress_timestamp()
                logging.info(
                    "fetch() with %d new records%s",
                    len(delta.updates),
                    (
                        ""
                        if progress_time is None
                        else f" @ {_timeformat(progress_time)}"
                    ),
                )
                next_checkpoint = delta.checkpoint
                self._fetch_progress(len(delta.updates), next_checkpoint)
                if checkpoint is not None:
                    prev_time = checkpoint.get_progress_timestamp()
                    if prev_time is not None and progress_time is not None:
                        assert prev_time <= progress_time, (
                            "checkpoint time rewound? ",
                            "This can indicate a serious ",
                            "problem with the API and checkpointing",
                        )
                checkpoint = next_checkpoint  # Only used for the rewind check

                store.merge(collab, delta)
                if self.has_hit_limits():
                    self.stderr("Hit limits, checkpointing")
                    break
                if self.should_checkpoint():
                    self._print_progress(checkpoint=True)
                    store.flush()
                    self.last_checkpoint_time = time.time()
            completed = True
        except Exception:
            self._stderr_current("failed to fetch!")
            logging.exception("Failed to fetch %s", collab.name)
            return False
        except KeyboardInterrupt:
            self._stderr_current("Interrupted, writing a checkpoint...")
            raise
        finally:
            store.flush()

        self._print_progress(done=completed)
        return True

    def _verify_store_and_checkpoint(
        self, store: FetchedStateStoreBase, collab: CollaborationConfigBase
    ) -> t.Optional[FetchCheckpointBase]:
        checkpoint = store.get_checkpoint(collab)

        if checkpoint is not None and checkpoint.is_stale():
            logging.info("checkpoint stale for %s, clearing", collab.name)
            store.clear(collab)
            return None

        return checkpoint

    def _fetch_progress(self, batch_size: int, checkpoint: FetchCheckpointBase) -> None:
        self.progress_fetched_count += batch_size
        self.total_fetched_count += batch_size
        progress_ts = checkpoint.get_progress_timestamp()
        if progress_ts is not None:
            self.last_update_time = progress_ts

        now = time.time()
        if now - self.last_update_printed >= self.PROGRESS_PRINT_INTERVAL_SEC:
            self.last_update_printed = now
            self._print_progress()

    def _stderr_current(self, msg: str) -> None:
        assert self.current_api and self.current_collab
        self.stderr(
            f"[{self.current_api}] {self.current_collab} - {msg}",
        )

    def _print_progress(self, *, done=False, checkpoint=False):
        processed = "Fetching..."
        if done:
            processed = "Up to date"
        elif checkpoint:
            processed = "Checkpointing"
        elif self.progress_fetched_count:
            processed = f"Downloaded {self.progress_fetched_count} updates"

        from_time = ""
        if self.last_update_time is not None:
            if self.last_update_time <= 0:
                from_time = "ages long past"
            elif self.last_update_time >= time.time() - 1:
                from_time = "moments ago"
            else:
                from_time = _timeformat(self.last_update_time)
            from_time = f" at {from_time}"

        self._stderr_current(f"{processed}{from_time}")


def _timeformat(timestamp: int) -> str:
    return datetime.datetime.fromtimestamp(timestamp).isoformat()



import pytest
from threatexchange.cli.tests.e2e_test_helper import (
    ThreatExchangeCLIE2eHelper,
    te_cli,
)


@pytest.fixture
def cli(
    te_cli: ThreatExchangeCLIE2eHelper,
) -> ThreatExchangeCLIE2eHelper:
    te_cli.COMMON_CALL_ARGS = ("config", "extensions")
    return te_cli


def test_extension_manipulate(cli: ThreatExchangeCLIE2eHelper) -> None:
    expected = [
        "Signal:",
        "  fake - FakeSignal",
        "Content:",
        "  fake - FakeContent",
        "Content:",
        "  fake - FakeSignalExchange",
    ]
    cli.assert_cli_output(["list"], "")
    cli.assert_cli_output(["add", "threatexchange.cli.tests.fake_extension"], expected)
    cli.assert_cli_output(
        ["list"],
        ["threatexchange.cli.tests.fake_extension"] + ["  " + e for e in expected],
    )
    cli.assert_cli_output(["list", "threatexchange.cli.tests.fake_extension"], expected)
    # Double add ok
    cli.assert_cli_output(["add", "threatexchange.cli.tests.fake_extension"], expected)

    cli.assert_cli_output(["remove", "threatexchange.cli.tests.fake_extension"], "")
    cli.assert_cli_output(["list"], "")

    # Double remove
    cli.assert_cli_usage_error(["remove", "threatexchange.cli.tests.fake_extension"])
    # Add noexist
    cli.assert_cli_usage_error(["list", "threatexchange.cli.tests.noexist"])


def test_extensions_in_other_cmds(cli: ThreatExchangeCLIE2eHelper) -> None:
    cli.cli_call("add", "threatexchange.cli.tests.fake_extension")
    cli.assert_cli_output(
        ["tx", "hash", "--signal-type=fake", "fake", "--", "lolol"], "fake fake"
    )
    cli.assert_cli_output(
        ["tx", "hash", "--signal-type=fake", "fake", "--", "lolol"], "fake fake"
    )
    cli.assert_cli_output(
        ["tx", "match", "--only-signal=fake", "fake", "--", "lolol"],
        "fake - (Sample Signals) INVESTIGATION_SEED",
    )



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import vpdq
import json
import typing as t
import pathlib
from dataclasses import dataclass

from threatexchange.signal_type.pdq.pdq_utils import PDQ_HEX_STR_LEN

QUALITY = "quality"
HASH = "hash"
TIMESTAMP = "timestamp"
VPDQ_TIMESTAMP_PRECISION = 3
VPDQ_QUALITY_THRESHOLD = 50
VPDQ_DISTANCE_THRESHOLD = 31
VPDQ_QUERY_MATCH_THRESHOLD_PERCENT = 80.0
VPDQ_INDEX_MATCH_THRESHOLD_PERCENT = 0.0


@dataclass
class VPDQMatchResult:
    """Data class for VPDQ match result"""

    query_match_percent: float = 0.0
    compared_match_percent: float = 0.0


@dataclass
class VpdqCompactFeature:
    """A VPDQ Feature with a subset of fields needed for matching"""

    pdq_hex: str
    quality: int
    timestamp: float

    def assert_valid(self) -> "VpdqCompactFeature":
        """Checks the bounds of all the elements, throws ValueError if invalid"""
        if len(self.pdq_hex) != PDQ_HEX_STR_LEN:
            raise ValueError("malformed pdq hash")
        int(self.pdq_hex, 16)  # For ValueError
        if not (0 <= self.quality <= 100):
            raise ValueError("invalid VPDQ quality")
        if self.timestamp < 0:
            raise ValueError("invalid timestamp")
        return self

    @classmethod
    def from_vpdq_feature(cls, feature: vpdq.VpdqFeature) -> "VpdqCompactFeature":
        return cls(feature.hex, feature.quality, feature.timestamp)

    @classmethod
    def from_str(cls, serialized: str) -> "VpdqCompactFeature":
        """Convert from a string back to the class - the inverse of __str__"""
        parts = serialized.split(",")
        try:
            pdq_hex, qual_str, time_str = parts  # Wrong count = ValueError
            return cls(pdq_hex, int(qual_str), float(time_str)).assert_valid()
        except ValueError:
            raise ValueError(f"invalid {cls.__name__} serialization: {serialized}")

    def __str__(self) -> str:
        return f"{self.pdq_hex},{self.quality},{self.timestamp:.{VPDQ_TIMESTAMP_PRECISION}}"


def hash_file_compact(
    filepath: str, seconds_per_hash: float = 1.0
) -> t.List[VpdqCompactFeature]:
    """Wrapper around computeHash to instead return compact features"""
    vpdq_hashes = vpdq.computeHash(str(filepath), seconds_per_hash=seconds_per_hash)
    return [VpdqCompactFeature.from_vpdq_feature(f) for f in vpdq_hashes]


def vpdq_to_json(
    vpdq_features: t.List[VpdqCompactFeature], *, indent: t.Optional[int] = None
) -> str:
    """Convert from VPDQ features to json object and return the json object as a str"""
    return json.dumps([str(f.assert_valid()) for f in vpdq_features], indent=indent)


def json_to_vpdq(json_str: str) -> t.List[VpdqCompactFeature]:
    """Load a str as a json object and convert from json object to VPDQ features"""
    return [VpdqCompactFeature.from_str(s) for s in json.loads(json_str or "[]")]


def dedupe(features: t.List[VpdqCompactFeature]) -> t.List[VpdqCompactFeature]:
    """Filter out the VPDQ feature with exact same hash in a list of VPDQ features

    Args:
        features

    Returns:
        List of VPDQ Features with unique features
    """
    unique_features = set()
    ret = []
    for h in features:
        if h.pdq_hex not in unique_features:
            ret.append(h)
            unique_features.add(h.pdq_hex)
    return ret


def quality_filter(
    features: t.List[VpdqCompactFeature], quality_tolerance: int
) -> t.List[VpdqCompactFeature]:
    """Filter VPDQ feature that has a quality lower than quality_tolerance

    Args:
        features
        quality_tolerance : If frames is this quality level then it will be ignored

    Returns:
        List of VPDQFeatures with quality higher than quality_tolerance
    """
    return [f for f in features if f.quality >= quality_tolerance]


def OLD_json_to_vpdq(json_str: str) -> t.List[vpdq.VpdqFeature]:
    """Load a str as a json object and convert from json object to VPDQ features"""
    if not json_str:
        return []
    features = []
    # VPDQ feature's timestamp is round to 3 decimals
    vpdq_json = json.loads(
        json_str, parse_float=lambda x: round(float(x), VPDQ_TIMESTAMP_PRECISION)
    )
    for frame_number, feature in vpdq_json.items():
        features.append(
            vpdq.VpdqFeature(
                feature[QUALITY], int(frame_number), feature[HASH], feature[TIMESTAMP]
            )
        )
    return features


def OLD_read_file_to_hash(
    input_hash_filename: t.Union[str, pathlib.Path]
) -> t.List[VpdqCompactFeature]:
    """Read hash file and return list of VPDQ features

    Args:
        Input hash file path

    Returns:
        VPDQ features from the hash file"""

    with open(input_hash_filename, "r") as file:
        return OLD_json_to_vpdq(file.read())


def OLD_dump_hash_to_file(
    output_hash_filename: t.Union[str, pathlib.Path],
    vpdq_features: t.List[VpdqCompactFeature],
) -> None:
    """Write list of VPDQ features to output hash file

    Args:
        Output hash file path
        VPDQ features write to the output file"""
    with open(output_hash_filename, "w") as file:
        file.write(vpdq_to_json(vpdq_features))


def prepare_vpdq_feature(
    signal_str: str, quality_tolerance: int
) -> t.List[VpdqCompactFeature]:
    """Convert signal_str to deduped and quality-filtered vdqp features

    Args:
    quality_tolerance : The quality tolerance of VPDQ Feature.
    If VPDQ Feature is below this quality level then it will not be added
    """
    features = json_to_vpdq(signal_str)
    return dedupe(quality_filter(features, quality_tolerance))



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from dataclasses import dataclass, field
import enum
import pathlib
import typing as t
from threatexchange.exchanges.collab_config import (
    CollaborationConfigBase,
    CollaborationConfigWithDefaults,
)
from threatexchange.exchanges.fetch_state import (
    FetchCheckpointBase,
    FetchDelta,
    FetchedSignalMetadata,
)

from threatexchange.extensions.manifest import ThreatExchangeExtensionManifest
from threatexchange.signal_type.signal_base import (
    FileHasher,
    SignalComparisonResult,
    SignalType,
    TrivialSignalTypeIndex,
)
from threatexchange.content_type.content_base import ContentType
from threatexchange.exchanges.signal_exchange_api import SignalExchangeAPI


class FakeContent(ContentType):
    pass


class FakeEnum(enum.Enum):
    OPTION_A = "a"
    OPTION_B = "b"
    CamelCase = "camels"
    lower_under = "under score"


@dataclass
class _FakeCollabConfigRequiredFields:
    an_int: int = field(metadata={"help": "Demonstrate int"})
    a_str: str = field(metadata={"help": "Demonstrate str"})
    a_list: t.List[str] = field(metadata={"help": "Demonstrate list"})
    a_set: t.Set[int] = field(metadata={"help": "Demonstrate set"})
    an_enum: FakeEnum = field(metadata={"help": "Demonstrate enum"})


@dataclass
class FakeCollabConfig(
    CollaborationConfigWithDefaults, _FakeCollabConfigRequiredFields
):
    optional: t.Optional[float] = field(
        default=None, metadata={"help": "Demonstrate optional float"}
    )


class FakeSignal(SignalType, FileHasher):
    @classmethod
    def get_content_types(cls) -> t.List[t.Type[ContentType]]:
        """Which content types this Signal applies to (usually just one)"""
        return [FakeContent]

    @classmethod
    def get_index_cls(cls) -> t.Type[TrivialSignalTypeIndex]:
        return TrivialSignalTypeIndex

    @classmethod
    def hash_from_file(cls, file: pathlib.Path) -> str:
        return "fake"  # A perfect hashing algorithm

    @classmethod
    def get_examples(cls) -> t.List[str]:
        return ["fake", "not fake"]


class FakeSignalExchange(
    SignalExchangeAPI[
        FakeCollabConfig, FetchCheckpointBase, FetchedSignalMetadata, str, str
    ]
):
    @classmethod
    def get_config_cls(cls) -> t.Type[FakeCollabConfig]:
        return FakeCollabConfig

    @classmethod
    def naive_convert_to_signal_type(
        cls,
        signal_types: t.Sequence[t.Type[SignalType]],
        collab: FakeCollabConfig,
        fetched: t.Mapping[str, str],
    ) -> t.Dict[t.Type[SignalType], t.Dict[str, FetchedSignalMetadata]]:
        return {}

    def fetch_iter(
        self,
        supported_signal_types: t.Sequence[t.Type[SignalType]],
        # None if fetching for the first time,
        # otherwise the previous FetchDelta returned
        checkpoint: t.Optional[FetchCheckpointBase],
    ) -> t.Iterator[FetchDelta[str, str, FetchCheckpointBase]]:
        return
        yield  # how to write an empty generator


TX_MANIFEST = ThreatExchangeExtensionManifest(
    signal_types=(FakeSignal,),
    content_types=(FakeContent,),
    apis=(FakeSignalExchange,),
)



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Hash command to convert content into signatures.
"""

import argparse
import pathlib
import typing as t
from threatexchange import common
from threatexchange.cli.cli_config import CLISettings
from threatexchange.cli.exceptions import CommandError
from threatexchange.content_type.content_base import ContentType

from threatexchange.signal_type.signal_base import FileHasher, SignalType
from threatexchange.cli import command_base
from threatexchange.cli.helpers import FlexFilesInputAction


class HashCommand(command_base.Command):
    """
    Take content and convert it into signatures (aka hashes)

    # Input
    You can pass in data to the command in a few different ways:
    (Note - you may not be able to hash text without an extension)
    ```
    # As an input file that contains one signal
    $ threatexchange hash photo my_photo.jpg

    # As stdin
    $ echo This is my cool text | threatexchange hash text -

    # Inline
    $ threatexchange hash text -- This is my cool text
    ```

    # Output
    <SignalType> <hash string>
    """

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:

        signal_types = [
            s for s in settings.get_all_signal_types() if issubclass(s, FileHasher)
        ]

        content_choices = sorted(s.get_name() for s in settings.get_all_content_types())
        signal_choices = sorted(
            s.get_name() for s in signal_types if issubclass(s, FileHasher)
        )
        ap.add_argument(
            "content_type",
            **common.argparse_choices_pre_type_kwargs(
                choices=content_choices,
                type=settings.get_content_type,
            ),
            help="what kind of content to hash",
        )

        ap.add_argument(
            "files",
            nargs=argparse.REMAINDER,
            action=FlexFilesInputAction,
            help="list of files, URLs, - for stdin, or -- to interpret remainder as a string",
        )

        ap.add_argument(
            "--signal-type",
            "-S",
            **common.argparse_choices_pre_type_kwargs(
                choices=signal_choices,
                type=settings.get_signal_type,
            ),
            help="only generate these signal types",
        )

    def __init__(
        self,
        content_type: t.Type[ContentType],
        signal_type: t.Optional[t.Type[SignalType]],
        files: t.List[pathlib.Path],
    ) -> None:
        self.content_type = content_type
        self.signal_type = signal_type

        self.files = files

    def execute(self, settings: CLISettings) -> None:
        hashers = [
            s
            for s in settings.get_signal_types_for_content(self.content_type)
            if issubclass(s, FileHasher)
        ]
        if self.signal_type is not None:
            if self.signal_type not in hashers:
                raise CommandError.user(
                    f"{self.signal_type.get_name()} "
                    f"does not apply to {self.content_type.get_name()}"
                )
            hashers = [self.signal_type]  # type: ignore  # can't detect intersection types

        for file in self.files:
            for hasher in hashers:
                hash_str = hasher.hash_from_file(file)
                if hash_str:
                    print(hasher.get_name(), hash_str)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import io
import pdqhash
import pathlib
import numpy as np
from PIL import Image
import typing as t


PDQOutput = t.Tuple[
    str, int
]  # hexadecimal representation of the Hash vector and a numerical quality value


def pdq_from_file(path: pathlib.Path) -> PDQOutput:
    """
    Given a path to a file return the PDQ Hash string in hex.
    Current tested against: jpg
    """
    img_pil = Image.open(path)
    image = _check_dimension_and_expand_if_needed(np.asarray(img_pil))
    return _pdq_from_numpy_array(image)


def pdq_from_bytes(file_bytes: bytes) -> PDQOutput:
    """
    For the bytestream from an image file, compute PDQ Hash and quality.
    """
    np_array = _check_dimension_and_expand_if_needed(
        np.asarray(Image.open(io.BytesIO(file_bytes)))
    )
    return _pdq_from_numpy_array(np_array)


def _pdq_from_numpy_array(array: np.ndarray) -> PDQOutput:
    hash_vector, quality = pdqhash.compute(array)

    bin_str = "".join([str(x) for x in hash_vector])

    # binary to hex using format string
    # '%0*' is for padding up to ceil(num_bits/4),
    # '%X' create a hex representation from the binary string's integer value
    hex_str = "%0*X" % ((len(bin_str) + 3) // 4, int(bin_str, 2))
    hex_str = hex_str.lower()

    return hex_str, quality


def _check_dimension_and_expand_if_needed(array: np.ndarray) -> np.ndarray:
    """
    Convert 2 dim array to the 3 dim 'pdqhash' expects
    (without this black and white images result in errors).
    """
    if array.ndim == 2:
        array = np.concatenate([array[..., np.newaxis]] * 3, axis=2)
    return array



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Local storage and configuration for the CLI.

The CLI and Hasher-Matcher-Actioner are roughly parallel, but this isn't a 
scalable service running on AWS. Instead, we have all of our state in
a file (likely ~/.threatexchange)
"""

from dataclasses import dataclass, field
import sys
import typing as t
import json
import pathlib
import logging

from dacite import WrongTypeError

from threatexchange.exchanges import collab_config
from threatexchange.exchanges.impl.stop_ncii_api import StopNCIICredentials
from threatexchange.exchanges.signal_exchange_api import (
    SignalExchangeAPI,
    TCollabConfig,
    TSignalExchangeAPICls,
)
from threatexchange.content_type import content_base
from threatexchange.exchanges import fetch_state
from threatexchange.exchanges.impl.static_sample import StaticSampleSignalExchangeAPI
from threatexchange.signal_type import signal_base
from threatexchange.interface_validation import FunctionalityMapping
from threatexchange.cli.cli_state import CliSimpleState, CliIndexStore
from threatexchange.utils import dataclass_json


CONFIG_FILENAME = "config.json"


@dataclass
class CLiConfig:
    """A place to store misc configuration for the CLI"""

    fb_threatexchange_api_token: t.Optional[str] = None
    ncmec_credentials: t.Optional[t.Tuple[str, str]] = None
    stop_ncii_keys: t.Optional[StopNCIICredentials] = None
    extensions: t.Set[str] = field(default_factory=set)
    # Every item needs a default for backwards compatibility


class CliState(collab_config.CollaborationConfigStoreBase):
    """
    A wrapper around stateful information stored for the CLI.

    Everything is just in a single directory (usually ~/.threatexchange).
    """

    def __init__(
        self,
        fetch_types: t.List[t.Union[SignalExchangeAPI, t.Type[SignalExchangeAPI]]],
        dir: pathlib.Path,
    ):
        self._dir = dir.expanduser()

        self._name_to_ctype = {ft.get_name(): ft.get_config_cls() for ft in fetch_types}

        self._cache: t.Optional[
            t.Dict[str, collab_config.CollaborationConfigBase]
        ] = None

        self._init_folders_if_needed()

    def _init_folders_if_needed(self):
        for d in (self.collab_dir, self.index_dir, self.fetched_state_dir):
            if not d.is_dir():
                d.mkdir(parents=True)
        cfg = self.config_file
        if not cfg.is_file():
            cfg.write_text("{}")

    @property
    def collab_dir(self) -> pathlib.Path:
        return self._dir / "collab_configs/"

    @property
    def fetched_state_dir(self) -> pathlib.Path:
        return self._dir / "fetched"

    @property
    def index_dir(self) -> pathlib.Path:
        return self._dir / "index/"

    @property
    def config_file(self) -> pathlib.Path:
        return self._dir / "config.json"

    def path_for_collab_config(
        self, config: collab_config.CollaborationConfigBase
    ) -> pathlib.Path:
        return self.collab_dir / f"{config.name}.json"

    def get_persistent_config(self) -> CLiConfig:
        return dataclass_json.dataclass_load_file(
            self.config_file, CLiConfig, default=CLiConfig()
        )

    def update_persistent_config(self, config: CLiConfig):
        dataclass_json.dataclass_dump_file(self.config_file, config)

    def dir_for_fetched_state(
        self,
        api: t.Type[SignalExchangeAPI],
    ) -> pathlib.Path:
        return self.fetched_state_dir / f"{api.get_name()}/"

    def get_collab_names_without_loading(self) -> t.List[str]:
        if self._cache is not None:
            return list(self._cache)
        return [str(p) for p in self.collab_dir.glob("*.json")]

    def get_all_collabs(self) -> t.List[collab_config.CollaborationConfigBase]:
        """
        Get all CollaborationConfigs, already resolved to the correct type
        """
        if self._cache is None:
            collab_dir = self.collab_dir

            ret = []
            for f in collab_dir.glob("*.json"):
                if not f.is_file():
                    logging.warning("Ignoring strange file in collab dir: %s", f)
                    continue
                with f.open() as fp:
                    try:
                        content = json.load(fp)
                    except json.JSONDecodeError:
                        logging.exception("Failed to parse collab config: %s", f)
                        continue
                    ctype = None
                    if isinstance(content, dict):
                        ctype = self._name_to_ctype.get(content.get("api"))  # type: ignore
                    if ctype is None:
                        logging.warning("Ignoring collab config of unknown type: %s", f)
                        continue
                try:
                    config = dataclass_json.dataclass_load_dict(content, ctype)
                    ret.append(config)
                except WrongTypeError:
                    logging.exception("Failed to parse collab config: %s", f)
            self._cache = {c.name: c for c in ret}
        return list(self._cache.values())

    def update_collab(self, collab: collab_config.CollaborationConfigBase) -> None:
        """Create or update a collaboration"""
        assert collab.api, "didn't set API?"
        path = self.path_for_collab_config(collab)
        dataclass_json.dataclass_dump_file(path, collab)

    def delete_collab(self, collab: collab_config.CollaborationConfigBase) -> None:
        """Delete a collaboration"""
        self.path_for_collab_config(collab).unlink(missing_ok=True)


@dataclass
class _SignalExchangeAccessor:
    """Convenience wrapper for operations on the SignalExchangeAPI"""

    _parent: "CLISettings"

    def get_all(self) -> t.ValuesView[TSignalExchangeAPICls]:
        return self._parent._mapping.exchange.api_by_name.values()

    def get_instance_for_collab(
        self, collab: TCollabConfig
    ) -> SignalExchangeAPI[
        TCollabConfig,
        fetch_state.FetchCheckpointBase,
        fetch_state.FetchedSignalMetadata,
        t.Any,
        t.Any,
    ]:
        api_cls = self._parent._mapping.exchange.api_by_name[collab.api]
        return api_cls.for_collab(collab)

    def __iter__(self) -> t.Iterator[TSignalExchangeAPICls]:
        yield from self.get_all()


@dataclass
class _FetchStoreAccessor:
    """Convenience wrapper for operations on the state"""

    _parent: "CLISettings"

    def empty(self) -> bool:
        """Return the collabs with stored state"""
        collabs = self._parent.get_all_collabs()
        return not any(
            collab for collab in collabs if self.get_for_collab(collab).exists(collab)
        )

    def get_for_api(self, api: t.Type[SignalExchangeAPI]) -> CliSimpleState:
        return CliSimpleState(api, self._parent._state.dir_for_fetched_state(api))

    def get_for_collab(
        self, collab: collab_config.CollaborationConfigBase
    ) -> CliSimpleState:
        return self.get_for_api(self._parent._mapping.exchange.api_by_name[collab.api])


class CLISettings:
    """
    A God object for all miscellanious persisted state to make the CLI work
    """

    def __init__(
        self,
        mapping: FunctionalityMapping,
        cli_state: CliState,
    ) -> None:
        self._mapping = mapping
        self._state = cli_state
        self._sample_message_printed = False
        self._config: t.Optional[CLiConfig] = None
        self.index = CliIndexStore(cli_state.index_dir)
        self.fetched_state = _FetchStoreAccessor(self)
        self.apis = _SignalExchangeAccessor(self)

    def get_persistent_config(self) -> CLiConfig:
        if self._config is None:
            self._config = self._state.get_persistent_config()
        return self._config

    def set_persistent_config(self, config: CLiConfig) -> None:
        self._state.update_persistent_config(config)
        self._config = config

    def get_all_content_types(self) -> t.List[t.Type[content_base.ContentType]]:
        return list(self._mapping.signal_and_content.content_by_name.values())

    def get_content_type(self, name: str) -> t.Type[content_base.ContentType]:
        return self._mapping.signal_and_content.content_by_name[name]

    def get_all_signal_types(self) -> t.List[t.Type[signal_base.SignalType]]:
        return list(self._mapping.signal_and_content.signal_type_by_name.values())

    def get_signal_type(self, name: str) -> t.Type[signal_base.SignalType]:
        return self._mapping.signal_and_content.signal_type_by_name[name]

    def get_signal_types_for_content(
        self, content_type: t.Type[content_base.ContentType]
    ) -> t.List[t.Type[signal_base.SignalType]]:
        return self._mapping.signal_and_content.signal_type_by_content[content_type]

    @property
    def in_demo_mode(self) -> bool:
        """Has no live collabs"""
        return not self._state.get_all_collabs()

    def get_all_collabs(
        self, *, default_to_sample: bool = True
    ) -> t.List[collab_config.CollaborationConfigBase]:
        if self.in_demo_mode and default_to_sample:
            return [self._get_sample_collab()]
        # Should this check whether the APIs are all valid?
        return self._state.get_all_collabs()

    def get_collab(
        self,
        name: str,
    ) -> t.Optional[collab_config.CollaborationConfigBase]:
        return self._state.get_collab(name)

    def _get_sample_collab(self) -> collab_config.CollaborationConfigBase:
        if not self._sample_message_printed:
            print(
                (
                    "Looks like you haven't set up a collaboration config, "
                    "so using the sample one against sample data"
                ),
                file=sys.stderr,
            )
            self._sample_message_printed = True
        return collab_config.CollaborationConfigBase(
            "Sample Signals", StaticSampleSignalExchangeAPI.get_name(), enabled=True
        )

    def get_collabs_for_api(
        self, api: SignalExchangeAPI
    ) -> t.List[collab_config.CollaborationConfigBase]:
        api_name = api.get_name()
        return [c for c in self.get_all_collabs() if c.api == api_name]



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Typed representations (dataclasses only) for interfacing with the
threatexchange API.
"""

from datetime import datetime
from dateutil.parser import parse
from dataclasses import dataclass


def _parse_datetime_from_iso_8601(datestr: str) -> datetime:
    """
    Parses strings representing date like 2019-05-20T16:44:47+0000 from the
    graph api into datetime objects.
    """
    return parse(datestr)


@dataclass
class ThreatPrivacyGroup:
    id: int
    name: str
    description: str
    members_can_see: bool
    members_can_use: bool
    threat_updates_enabled: bool
    last_updated: datetime

    def __eq__(self, other) -> bool:
        return self.id == other.id

    def __hash__(self) -> int:
        return hash(self.id)

    @classmethod
    def from_graph_api_dict(cls, d: dict) -> "ThreatPrivacyGroup":
        return cls(
            d["id"],
            d["name"],
            d["description"],
            bool(d["members_can_see"]),
            bool(d["members_can_use"]),
            bool(d["threat_updates_enabled"]),
            _parse_datetime_from_iso_8601(d["last_updated"]),
        )



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from argparse import ArgumentParser
import collections
import csv
import sys
import typing as t
import logging
from threatexchange.cli.exceptions import CommandError


from threatexchange.signal_type.signal_base import SignalType
from threatexchange.cli.cli_config import CLISettings
from threatexchange.content_type.content_base import ContentType
from threatexchange.exchanges.collab_config import CollaborationConfigBase
from threatexchange.exchanges.fetch_state import FetchedSignalMetadata

from threatexchange.cli import command_base
from threatexchange import common


class DatasetCommand(command_base.Command):
    """
    Introspect fetched data.

    Can print out contents in simple formats
    (ideal for sending to another system), or regenerate
    index files (ideal if distributing indices for some reason)

    Example commands:

    ```
    # Show total size of known signals
    $ threatexchange dataset

    # Show the size of one collaboration
    $ threatexchange dataset -c 'Sample Data'

    # Generate a simple CSV file for one collaboration for ease of distribution
    $ threatexchange dataset -c 'Sample Data' -P --csv > collab.csv
    ```
    """

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: ArgumentParser) -> None:
        actions = ap.add_mutually_exclusive_group()
        actions.add_argument(
            "--rebuild-indices",
            "-r",
            action="store_true",
            help="rebuild indices from fetched data",
        )
        actions.add_argument(
            "--clear-indices",
            "-X",
            action="store_true",
            help="clear all indices",
        )
        actions.add_argument(
            "--signal-summary",
            action="store_true",
            help="print summary in terms of signals (default action)",
        )
        actions.add_argument(
            "--print-signals",
            "-P",
            action="store_true",
            help="print signals to screen",
        )

        type_selector = ap.add_mutually_exclusive_group()
        type_selector.add_argument(
            "--only-signals",
            "-s",
            nargs="+",
            default=[],
            type=common.argparse_choices_pre_type(
                choices=[s.get_name() for s in settings.get_all_signal_types()],
                type=settings.get_signal_type,
            ),
            metavar="SIGNAL_TYPE",
            help="only use signals of this type",
        )
        type_selector.add_argument(
            "--only-content",
            "-C",
            nargs="+",
            default=[],
            type=common.argparse_choices_pre_type(
                choices=[s.get_name() for s in settings.get_all_content_types()],
                type=settings.get_content_type,
            ),
            metavar="CONTENT_TYPE",
            help="only use signals for these content types",
        )
        ap.add_argument(
            "--print-zeroes",
            "-z",
            action="store_true",
            help="[--signal-summary] print counts of 0",
        )
        ap.add_argument(
            "--only-collabs",
            "-c",
            nargs="+",
            default=[],
            metavar="NAME",
            help="[-S|-P] only use items with this tag",
        )
        ap.add_argument(
            "--only-tags",
            "-t",
            nargs="+",
            default=[],
            metavar="STR",
            help="[-S|-P] only use items with these tags",
        )
        csv_mutual_group = ap.add_mutually_exclusive_group()
        csv_mutual_group.add_argument(
            "--print-signals-only",
            "-S",
            action="store_true",
            help="[-P] print type and signal only, no metadata",
        )
        csv_mutual_group.add_argument(
            "--csv",
            action="store_true",
            help="[-P] print in csv format (including header)",
        )

    def __init__(
        # These all have defaults to make it easier to call
        # only for rebuild
        self,
        # Mode
        clear_indices: bool = False,
        rebuild_indices: bool = False,
        signal_summary: bool = False,
        print_signals: bool = False,
        # Signal selectors
        only_collabs: t.Sequence[str] = (),
        only_signals: t.Sequence[t.Type[SignalType]] = (),
        only_content: t.Sequence[t.Type[ContentType]] = (),
        only_tags: t.Sequence[str] = (),
        # Print stuff
        print_zeroes: bool = False,
        print_signals_only: bool = False,
        csv: bool = False,
    ) -> None:
        self.clear_indices = clear_indices
        self.rebuild_indices = rebuild_indices
        self.print_signals = print_signals
        self.signal_summary = signal_summary or not (
            print_signals or rebuild_indices or clear_indices
        )

        self.only_collabs = set(only_collabs)
        self.only_signals = set(only_signals)
        self.only_content = set(only_content)
        self.only_tags = set(only_tags)

        self.print_zeroes = print_zeroes
        self.print_signals_only = print_signals_only
        self.csv = csv

    def execute(self, settings: CLISettings) -> None:
        if settings.fetched_state.empty():
            if not settings.in_demo_mode:
                raise CommandError("No stored state available. Do you need to fetch?")
            # FetchCommand currently imports dataset for index build, so inline for circular import
            from threatexchange.cli.fetch_cmd import FetchCommand

            self.stderr("You haven't fetched any state, so we'll call `fetch` for you!")
            FetchCommand().execute(settings)
        # Maybe consider subcommands?
        if self.clear_indices:
            self.execute_clear_indices(settings)
        elif self.rebuild_indices:
            self.execute_generate_indices(settings)
        elif self.print_signals:
            self.execute_print_signals(settings)
        else:
            assert self.signal_summary
            self.execute_print_summary(settings)

    def get_signal_types(self, settings: CLISettings) -> t.Set[t.Type[SignalType]]:
        signal_types = self.only_signals or settings.get_all_signal_types()
        if self.only_content:
            signal_types = [
                s
                for s in signal_types
                if any(c in self.only_content for c in s.get_content_types())
            ]
        return set(signal_types)

    def get_collabs(self, settings: CLISettings) -> t.List[CollaborationConfigBase]:
        collabs = [c for c in settings.get_all_collabs() if c.enabled]
        if self.only_collabs:
            collabs = [c for c in collabs if c.name in self.only_collabs]
        return collabs

    def get_signals(
        self, settings: CLISettings, signal_types: t.Iterable[t.Type[SignalType]]
    ) -> t.Dict[
        t.Type[SignalType], t.Dict[str, t.List[t.Tuple[str, FetchedSignalMetadata]]]
    ]:
        """
        Get signals grouped by type => hash => collab name: metadata
        """
        collabs = self.get_collabs(settings)

        collab_by_api: t.Dict[str, t.List[CollaborationConfigBase]] = {}
        for collab_config in collabs:
            collab_by_api.setdefault(collab_config.api, []).append(collab_config)
        by_type = {}
        for s_type in signal_types:
            by_signal: t.Dict[
                str,
                t.List[t.Tuple[str, FetchedSignalMetadata]],
            ] = {}
            for collabs_for_store in collab_by_api.values():
                store = settings.fetched_state.get_for_collab(collabs_for_store[0])
                by_collab = store.get_for_signal_type(collabs_for_store, s_type)
                for collab, signals in by_collab.items():
                    for signal, record in signals.items():
                        if self.only_tags:
                            for opinion in record.get_as_opinions():
                                if any(t in self.only_tags for t in opinion.tags):
                                    break
                            else:
                                continue
                        by_signal.setdefault(signal, []).append((collab, record))
            by_type[s_type] = by_signal
        return by_type

    def execute_print_summary(self, settings: CLISettings):
        signals = self.get_signals(settings, self.get_signal_types(settings))
        by_type: t.Dict[str, int] = collections.Counter()
        for s_type, type_signals in signals.items():
            by_type[s_type.get_name()] += len(type_signals)
        for s_name, count in sorted(by_type.items(), key=lambda i: (-i[1], i[0])):
            if count or self.print_zeroes:
                print(f"{s_name}: {count}")

    def execute_print_signals(self, settings):
        signals = self.get_signals(settings, self.get_signal_types(settings))

        print_fn = self._print_stdout
        if self.csv:
            csv_writer = csv.DictWriter(
                sys.stdout, ["signal_type", "signal_str", "collab", "category", "tags"]
            )
            csv_writer.writeheader()
            print_fn = lambda *args: self._print_csv(csv_writer, *args)

        for signal_type, per_signal in signals.items():
            for signal_str, collab_signals in per_signal.items():
                if self.print_signals_only:
                    print_fn("", signal_type, signal_str, None)
                    continue
                for collab_name, metadata in collab_signals:
                    print_fn(collab_name, signal_type, signal_str, metadata)

    def execute_clear_indices(self, settings: CLISettings) -> None:
        only_signals = None
        if self.only_signals or self.only_content:
            only_signals = self.get_signal_types(settings)
        settings.index.clear(only_signals)

    def execute_generate_indices(self, settings: CLISettings) -> None:
        signal_types = self.get_signal_types(settings)

        for s_type in signal_types:
            index_cls = s_type.get_index_cls()
            signal_by_type = self.get_signals(settings, [s_type])
            signals = signal_by_type.get(s_type, {})
            if not signals:
                logging.info("No signals for %s", s_type.__name__)
                settings.index.clear([s_type])
                continue

            self.stderr(
                "Building index for",
                s_type.get_name(),
                f"with {len(signals)} signals...",
            )
            index = index_cls.build(signals.items())
            settings.index.store(s_type, index)
            self.stderr(f"Index for {s_type.get_name()} ready")

    def _print_stdout(
        self,
        collab_name: str,
        signal_type: SignalType,
        signal_str: str,
        metadata: t.Optional[FetchedSignalMetadata],
    ) -> None:
        if len(self.only_signals) != 1:
            print(signal_type.get_name(), end=" ")

        print(signal_str, end="")

        if not self.print_signals_only:
            if len(self.only_collabs) != 1:
                print("", repr(collab_name), end="")
            print("", metadata, end="")
        print()  # Complete line

    def _print_csv(
        self,
        csvwriter: csv.DictWriter,
        collab_name: str,
        signal_type: SignalType,
        signal_str: str,
        metadata: t.Optional[FetchedSignalMetadata],
    ) -> None:
        assert metadata is not None
        agg = metadata.get_as_aggregate_opinion()
        csvwriter.writerow(
            {
                "signal_type": signal_type.get_name(),
                "signal_str": signal_str,
                "collab": collab_name,
                "category": agg.category.name,
                "tags": " ".join(agg.tags),
            }
        )



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import unittest
import binascii
import random

from threatexchange.signal_type.pdq.pdq_utils import *
from threatexchange.signal_type.pdq.signal import PdqSignal
from threatexchange.tests.hashing.utils import (
    get_zero_hash,
    get_random_hash,
    get_similar_hash,
)


test_hashes = [
    "0000000000000000000000000000000000000000000000000000000000000000",
    "0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f",
    "f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0",
    "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
]


class TestPDQUtils(unittest.TestCase):
    def test_hex_to_binary(self):
        for test_hash in test_hashes:
            bin_hash = hex_to_binary_str(test_hash)
            self.assertEqual(len(bin_hash), BITS_IN_PDQ)
        self.assertEqual(hex_to_binary_str(test_hashes[0]), "0" * BITS_IN_PDQ)
        self.assertEqual(hex_to_binary_str(test_hashes[-1]), "1" * BITS_IN_PDQ)
        self.assertEqual(
            hex_to_binary_str(test_hashes[1]), "00001111" * (BITS_IN_PDQ // 8)
        )

    def test_binary_to_hex(self):
        for test_hash in test_hashes:
            bin_hash = hex_to_binary_str(test_hash)
            self.assertEqual(binary_str_to_hex(bin_hash), test_hash)

    def test_get_random_hash(self):
        for i in range(100):
            random_hash = get_random_hash()
            self.assertTrue(PdqSignal.validate_signal_str(random_hash))

    def test_get_similar_hash(self):
        for i in range(10):
            random_dist = random.randint(0, BITS_IN_PDQ)
            random_hash = get_random_hash()
            similar_hash = get_similar_hash(random_hash, random_dist)
            self.assertEqual(simple_distance(similar_hash, random_hash), random_dist)
        random_hash = get_random_hash()
        self.assertEqual(get_similar_hash(random_hash, 0), random_hash)
        # Max dist
        get_similar_hash(get_random_hash(), BITS_IN_PDQ)
        # Invalid dist
        self.assertRaises(
            ValueError, get_similar_hash, get_random_hash(), BITS_IN_PDQ + 1
        )
        self.assertRaises(ValueError, get_similar_hash, get_random_hash(), -1)

    def test_distance_binary(self):
        self.assertEqual(
            simple_distance_binary("0" * BITS_IN_PDQ, "1" * BITS_IN_PDQ), BITS_IN_PDQ
        )
        self.assertEqual(
            simple_distance_binary("0" * BITS_IN_PDQ, "0" * BITS_IN_PDQ), 0
        )
        self.assertEqual(
            simple_distance_binary(
                "0" * BITS_IN_PDQ, hex_to_binary_str(test_hashes[1])
            ),
            BITS_IN_PDQ // 2,
        )

    def test_distance(self):
        self.assertEqual(simple_distance(test_hashes[1], test_hashes[1]), 0)
        self.assertEqual(simple_distance(test_hashes[1], test_hashes[2]), BITS_IN_PDQ)
        self.assertEqual(
            simple_distance(test_hashes[1], test_hashes[3]), BITS_IN_PDQ // 2
        )

    def test_match_threshold(self):
        self.assertFalse(pdq_match(test_hashes[0], test_hashes[1], threshold=31))
        self.assertTrue(
            pdq_match(test_hashes[0], test_hashes[1], threshold=BITS_IN_PDQ)
        )
        self.assertTrue(pdq_match(test_hashes[0], test_hashes[0], threshold=0))
        self.assertFalse(pdq_match(test_hashes[0], test_hashes[1], threshold=0))
        self.assertTrue(pdq_match(test_hashes[1], test_hashes[3], BITS_IN_PDQ // 2))
        self.assertFalse(
            pdq_match(test_hashes[1], test_hashes[3], BITS_IN_PDQ // 2 - 1)
        )


if __name__ == "__main__":
    unittest.main()



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved
import typing as t
from threatexchange.extensions.vpdq.vpdq_util import VpdqCompactFeature
from threatexchange.tests.hashing.utils import get_random_hash


def get_random_vpdq_features(
    frame_count: int, seconds_per_frame: float = 1.0, quality: int = 100
) -> t.List[VpdqCompactFeature]:
    """Return a List which contains frame_count random VPDQ features with same quality and each feature's time stamp differs by seconds_per_frame"""
    return pdq_hashes_to_vpdq_features(
        [get_random_hash() for _ in range(frame_count)], seconds_per_frame, quality
    )


def pdq_hashes_to_vpdq_features(
    pdq_hashes: t.List[str], seconds_per_frame: float = 1.0, quality: int = 100
) -> t.List[VpdqCompactFeature]:
    """Return a List of VPDQ features generated from pdq_hashes with same quality and each feature's time stamp differs by seconds_per_frame"""
    return [
        VpdqCompactFeature(pdq_hash, quality, i * seconds_per_frame)
        for i, pdq_hash in enumerate(pdq_hashes)
    ]



##!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Common helpers and libraries for the all-in-one command.

Strongly consider moving classes and functions out of this file if it starts
to fill up.
"""

import argparse
import typing as t
import sys

from threatexchange import common
from threatexchange.cli.cli_config import CLISettings
from threatexchange.cli.exceptions import CommandError


class Command:
    """
    Simple wrapper around setting up commands for an argparse-based cli.
    """

    @classmethod
    def add_command_to_subparser(
        cls, settings: CLISettings, subparsers
    ) -> argparse.ArgumentParser:
        """
        Shortcut for adding the command to the parser.

        Propbably don't override.
        """
        command_ap = subparsers.add_parser(
            cls.get_name(),
            description=cls.get_description(),
            help=cls.get_help(),
            formatter_class=argparse.RawDescriptionHelpFormatter,
        )
        command_ap.set_defaults(command_cls=cls)
        cls.init_argparse(settings, command_ap)
        return command_ap

    @classmethod
    def init_argparse(
        cls, settings: CLISettings, argparse: argparse.ArgumentParser
    ) -> None:
        """
        Program the command subparser for __init__

        Your argument names should match the argument names in __init__.
        Be careful of collisions with the top level arg names from main.py
        """
        pass

    @classmethod
    def get_name(cls) -> str:
        """The display name of the command"""
        return common.class_name_to_human_name(cls.__name__, "Command").replace(
            "_", "-"
        )

    @classmethod
    def get_description(cls) -> str:
        """The long help of the command"""
        return cls.__doc__ or ""

    @classmethod
    def get_help(cls) -> str:
        """The short help of the command"""
        line = cls.get_description().strip().partition("\n")[0]
        # Good luck debugging this! Slightly reformat short description
        # (toplevel --help)
        if line and line[0].isupper():
            first_word, sp, rem = line.partition(" ")
            line = f"{first_word.lower()}{sp}{rem}"
        if line and line[-1] == ".":
            line = line[:-1]
        return line

    @staticmethod
    def stderr(*args, **kwargs) -> None:
        """Convenience accessor to stderr"""
        print(*args, file=sys.stderr, **kwargs)

    def execute(self, settings: CLISettings) -> None:
        raise NotImplementedError


class CommandWithSubcommands(Command):
    _SUBCOMMANDS: t.List[t.Type[Command]] = []

    @classmethod
    def add_command_to_subparser(
        cls, settings: CLISettings, subparsers
    ) -> argparse.ArgumentParser:
        command_ap = super().add_command_to_subparser(settings, subparsers)
        sub_subparsers = command_ap.add_subparsers()

        for command in cls._SUBCOMMANDS:
            command.add_command_to_subparser(settings, sub_subparsers)

        return command_ap

    def execute(self, settings: CLISettings) -> None:
        raise CommandError("subcommand is required", 2)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

import unittest
import binascii
import pickle
import numpy

from threatexchange.signal_type.pdq.pdq_faiss_matcher import (
    PDQFlatHashIndex,
    PDQMultiHashIndex,
)

test_hashes = [
    "0000000000000000000000000000000000000000000000000000000000000000",
    "000000000000000000000000000000000000000000000000000000000000ffff",
    "0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f",
    "f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0f0",
    "ffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffffff",
]

MAX_UNSIGNED_INT64 = numpy.iinfo(numpy.uint64).max


class MixinTests:
    class PDQHashIndexCommonTests(unittest.TestCase):
        index = None

        def assertEqualPDQHashSearchResults(self, result, expected):
            self.assertEqual(
                len(result), len(expected), "search results not of expected length"
            )
            for (r, e) in zip(result, expected):
                self.assertCountEqual(r, e)

        def test_search_index_for_exact_matches(self):
            query = test_hashes[:1]
            result = self.index.search(query, 0)
            self.assertEqualPDQHashSearchResults(result, [[q] for q in query])

        def test_search_index_for_near_matches(self):
            query = test_hashes[:1]
            result = self.index.search(query, 16)
            self.assertEqualPDQHashSearchResults(result, [test_hashes[:2]])

        def test_search_index_for_far_match(self):
            query = test_hashes[:1]
            result = self.index.search(query, 128)
            self.assertEqualPDQHashSearchResults(result, [test_hashes[:-1]])

        def test_search_index_multiple_queries(self):
            query = test_hashes
            result = self.index.search(query, 0)
            self.assertEqualPDQHashSearchResults(result, [[h] for h in test_hashes])

        def test_search_index_with_no_match(self):
            query = ["aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa"]
            result = self.index.search(query, 0)
            self.assertEqualPDQHashSearchResults(result, [[]])

        def test_search_index_with_multiple_no_matches(self):
            query = [
                "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
                "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
            ]
            result = self.index.search(query, 0)
            self.assertEqualPDQHashSearchResults(result, [[], []])

        def test_search_index_with_multiple_mixed_results(self):
            query = [
                "aaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaaa",
                "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb",
                test_hashes[-1],
            ]
            result = self.index.search(query, 0)
            self.assertEqualPDQHashSearchResults(result, [[], [], [test_hashes[-1]]])

        def test_supports_pickling(self):
            pickled_data = pickle.dumps(self.index)
            assert (
                pickled_data != None
            ), "index does not support pickling to a data stream"

            reconstructed_index = pickle.loads(pickled_data)
            assert (
                reconstructed_index != None
            ), "index does not support unpickling from data stream"
            assert (
                reconstructed_index.faiss_index != self.index.faiss_index
            ), "unpickling should create it's own faiss index in memory"

            query = [test_hashes[0]]
            result = reconstructed_index.search(query, 0)
            self.assertEqualPDQHashSearchResults(result, [[test_hashes[0]]])


class TestPDQFlatHashIndex(MixinTests.PDQHashIndexCommonTests, unittest.TestCase):
    def setUp(self):
        self.index = PDQFlatHashIndex()
        self.index.add(test_hashes, range(0, len(test_hashes)))

    def test_create_faiss_index_from_hashes(self):
        assert type(self.index) is PDQFlatHashIndex
        hashes_as_bytes = [binascii.unhexlify(h) for h in test_hashes]

        assert self.index.faiss_index is not None
        assert self.index.faiss_index.ntotal == len(test_hashes)

    def test_hash_at(self):
        assert test_hashes[2] == self.index.hash_at(2)

    def test_search_index_return_ids(self):
        query = test_hashes[:2]
        results = self.index.search(query, 16, return_as_ids=True)
        self.assertEqualPDQHashSearchResults(results, [[0, 1], [0, 1]])


class TestPDQFlatHashIndexWithCustomIds(
    MixinTests.PDQHashIndexCommonTests, unittest.TestCase
):

    custom_ids = [MAX_UNSIGNED_INT64 - i for i in range(len(test_hashes))]

    def setUp(self):
        self.index = PDQFlatHashIndex()
        self.index.add(test_hashes, self.custom_ids)

    def test_create_faiss_index_from_hashes(self):
        assert type(self.index) is PDQFlatHashIndex
        hashes_as_bytes = [binascii.unhexlify(h) for h in test_hashes]

        assert self.index.faiss_index is not None
        assert self.index.faiss_index.ntotal == len(test_hashes)

    def test_hash_at(self):
        assert test_hashes[2] == self.index.hash_at(self.custom_ids[2])

    def test_search_index_return_ids(self):
        query = test_hashes[:2]
        results = self.index.search(query, 16, return_as_ids=True)
        self.assertEqualPDQHashSearchResults(
            results, [self.custom_ids[:2], self.custom_ids[:2]]
        )


class TestPDQMultiHashIndex(MixinTests.PDQHashIndexCommonTests, unittest.TestCase):
    def setUp(self):
        self.index = PDQMultiHashIndex()
        self.index.add(test_hashes, range(0, len(test_hashes)))

    def test_create_faiss_index_from_hashes(self):
        assert type(self.index) is PDQMultiHashIndex
        hashes_as_bytes = [binascii.unhexlify(h) for h in test_hashes]

        assert self.index.faiss_index is not None
        assert self.index.faiss_index.ntotal == len(test_hashes)

    def test_hash_at(self):
        assert test_hashes[2] == self.index.hash_at(2)

    def test_search_index_return_ids(self):
        query = test_hashes[:2]
        results = self.index.search(query, 16, return_as_ids=True)
        self.assertEqualPDQHashSearchResults(results, [[0, 1], [0, 1]])


class TestPDQMultiHashIndexWithCustomIds(
    MixinTests.PDQHashIndexCommonTests, unittest.TestCase
):
    custom_ids = [MAX_UNSIGNED_INT64 - i for i in range(len(test_hashes))]

    def setUp(self):
        self.index = PDQMultiHashIndex()
        self.index.add(test_hashes, custom_ids=self.custom_ids)

    def test_create_faiss_index_from_hashes(self):
        assert type(self.index) is PDQMultiHashIndex
        hashes_as_bytes = [binascii.unhexlify(h) for h in test_hashes]

        assert self.index.faiss_index is not None
        assert self.index.faiss_index.ntotal == len(test_hashes)

    def test_hash_at(self):
        assert test_hashes[2] == self.index.hash_at(self.custom_ids[2])

    def test_search_index_return_ids(self):
        query = test_hashes[:2]
        results = self.index.search(query, 16, return_as_ids=True)
        self.assertEqualPDQHashSearchResults(
            results, [self.custom_ids[:2], self.custom_ids[:2]]
        )


if __name__ == "__main__":
    unittest.main()



from threatexchange.cli.tests.e2e_test_helper import ThreatExchangeCLIE2eTest
from threatexchange.signal_type.raw_text import RawTextSignal


class SampleDataE2ETest(ThreatExchangeCLIE2eTest):
    """
    The CLI should demonstrate functionality with only sample data.
    """

    MATCHES_ONE = RawTextSignal.get_examples()[0]
    MATCHES_TWO = RawTextSignal.get_examples()[-1]

    def test_direct_to_match(self):
        """The classic first use of the CLI"""
        self.assert_cli_output(
            ("match", "text", "--", self.MATCHES_ONE),
            "raw_text 100% (Sample Signals) INVESTIGATION_SEED",
        )

    def test_sequential_to_match(self):
        """Same as the classic first use, broken to component parts"""
        self.cli_call("fetch")
        self.cli_call("dataset", "-r")
        self.assert_cli_output(
            ("match", "text", "--", self.MATCHES_ONE),
            "raw_text 100% (Sample Signals) INVESTIGATION_SEED",
        )

    def test_multiple_match(self):
        """Matches should be able to hit multiple types"""
        self.assert_cli_output(
            ("match", "text", "--", self.MATCHES_TWO),
            "\n".join(
                (
                    "raw_text 100% (Sample Signals) INVESTIGATION_SEED",
                    "trend_query - (Sample Signals) INVESTIGATION_SEED",
                ),
            ),
        )

    def test_no_labeling(self):
        """You can't label the sample dataset"""
        self.assert_cli_usage_error(("label", "Sample Signals", "text", "foo"))



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Provides a list of which functionality an expansion provies.

See the README.md in this directory for more information.

tl;dr: Have a module with TX_MANIFEST that is assigned this class.
"""


from dataclasses import dataclass
import importlib
import typing as t

from threatexchange.exchanges.signal_exchange_api import SignalExchangeAPI
from threatexchange.signal_type.signal_base import SignalType
from threatexchange.content_type.content_base import ContentType


@dataclass
class ThreatExchangeExtensionManifest:
    """
    Class with a list of expansion contents.

    Assign a variable named TX_MANIFEST with this class populated with the
    contents of your extension to make it loadable by ThreatExchange.
    """

    signal_types: t.Tuple[t.Type[SignalType], ...] = ()
    content_types: t.Tuple[t.Type[ContentType], ...] = ()
    apis: t.Tuple[t.Type[SignalExchangeAPI], ...] = ()

    @classmethod
    def load_from_module_name(
        cls, module_name: str
    ) -> "ThreatExchangeExtensionManifest":
        """Following the expected conventions, load an extension"""
        try:
            module = importlib.import_module(module_name)
        except (ImportError, ValueError):
            raise ValueError(f"No such module '{module_name}'")

        try:
            manifest = module.TX_MANIFEST
        except AttributeError:
            raise ValueError(f"Module is missing TX_MANIFEST")

        if not isinstance(manifest, cls):
            raise ValueError(f"TX_MANIFEST is not a {cls.__name__}!")
        return manifest



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Wrapper around the video content type.
"""

from .content_base import ContentType


class PhotoContent(ContentType):
    """
    Content that contains static visual imagery.

    Examples might be:
      * jpeg
      * png
      * gif (non-animated)
      * frames from videos
      * thumbnails of videos
    """



#!/usr/bin/env python
# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Match command for parsing simple data sources against the dataset.
"""

import argparse
import logging
import pathlib
import typing as t


from threatexchange import common
from threatexchange.cli.fetch_cmd import FetchCommand
from threatexchange.cli.helpers import FlexFilesInputAction
from threatexchange.exchanges.fetch_state import FetchedSignalMetadata

from threatexchange.signal_type.index import IndexMatch, SignalTypeIndex
from threatexchange.cli.exceptions import CommandError
from threatexchange.signal_type.signal_base import BytesHasher, SignalType
from threatexchange.cli.cli_config import CLISettings
from threatexchange.content_type.content_base import ContentType

from threatexchange.signal_type.signal_base import MatchesStr, TextHasher, FileHasher
from threatexchange.cli import command_base


TMatcher = t.Callable[[pathlib.Path], t.List[IndexMatch]]


class MatchCommand(command_base.Command):
    """
    Match content to fetched signals

    Runs the given content through applicable SignalTypes, and compare it
    with previously fetched signals stored in the index files. Only
    SignalTypes supported by the CLI and your extensions can be matched,
    even if you can fetch them. Any matches will be printed to screen.

    # Input
    You can pass in data to the command in a few different ways:
    ```
    # As an input file that contains one signal
    $ threatexchange match photo my_photo.jpg

    # As stdin
    $ echo This is my cool text | threatexchange match text -

    # Inline
    $ threatexchange match text -- This is my cool text
    ```

    # Output
    The output of this command is in the following format:

    <signal type> - (<Collab Name>) <opinion category> <label1>,<label2>,...

    The category is key to understanding what you might want to do with a match.
    Here's an explanation of the categories. Opinion categories:
    * POSITIVE_CLASS:
      All contributors of this signal believe that matching content should belong
    * INVESTIGATION_SEED:
      This content needs manual investigation to confirm or fanout to find the
      content that fits the collaboration
    * DISPUTED:
      Some members have said that this signal can be used to find content
      that fits the collaboration, but others have said it matches content
      that does not belong in the collaboration.
    * NEGATIVE_CLASS:
      Members have said content that matches does not belong in the
      collaboration, matches the wrong content on their platform, or that this
      is informational content that shouldn't be treated the same as
      POSITIVE_CLASS content.
    """

    @classmethod
    def init_argparse(cls, settings: CLISettings, ap: argparse.ArgumentParser) -> None:

        ap.add_argument(
            "content_type",
            **common.argparse_choices_pre_type_kwargs(
                [c.get_name() for c in settings.get_all_content_types()],
                settings.get_content_type,
            ),
            help="what kind of content to match",
        )

        ap.add_argument(
            "--only-signal",
            "-S",
            **common.argparse_choices_pre_type_kwargs(
                [s.get_name() for s in settings.get_all_signal_types()],
                settings.get_signal_type,
            ),
            help="limit to this signal type",
        )

        ap.add_argument(
            "--hashes",
            "-H",
            action="store_true",
            help=("force input to be interpreted as signals for the given signal type"),
        )

        ap.add_argument(
            "files",
            nargs=argparse.REMAINDER,
            action=FlexFilesInputAction,
            help="list of files or -- to interpret remainder as a string",
        )

        ap.add_argument(
            "--show-false-positives",
            action="store_true",
            help="show matches even if you've marked them false_positive",
        )

        ap.add_argument(
            "--hide-disputed",
            action="store_true",
            help="hide matches if someone has disputed them",
        )
        ap.add_argument(
            "--all",
            "-A",
            action="store_true",
            help="show all matches, not just one per collaboration",
        )

    def __init__(
        self,
        content_type: t.Type[ContentType],
        only_signal: t.Optional[t.Type[SignalType]],
        hashes: bool,
        files: t.List[pathlib.Path],
        show_false_positives: bool,
        hide_disputed: bool,
        all: bool,
    ) -> None:
        self.content_type = content_type
        self.only_signal = only_signal
        self.as_hashes = hashes
        self.show_false_positives = show_false_positives
        self.hide_disputed = hide_disputed
        self.files = files
        self.all = all

        if only_signal and content_type not in only_signal.get_content_types():
            raise CommandError(
                f"{only_signal.get_name()} does not "
                f"apply to {content_type.get_name()}",
                2,
            )

    def execute(self, settings: CLISettings) -> None:
        if not settings.index.list():
            if not settings.in_demo_mode:
                raise CommandError("No indices available. Do you need to fetch?")
            self.stderr("You haven't built any indices, so we'll call `fetch` for you!")
            FetchCommand().execute(settings)

        signal_types = settings.get_signal_types_for_content(self.content_type)

        if self.only_signal:
            signal_types = [self.only_signal]
        types: t.Tuple[type, ...] = (FileHasher, MatchesStr)
        if self.as_hashes:
            types = (BytesHasher, TextHasher, FileHasher)
        signal_types = [s for s in signal_types if issubclass(s, types)]
        if self.as_hashes and len(signal_types) > 1:
            raise CommandError(
                f"Error: '{self.content_type.get_name()}' supports more than one SignalType."
                " for '--hashes' also use '--only-signal' to specify one of "
                f"{[s.get_name() for s in signal_types]}",
                2,
            )

        logging.info(
            "Signal types that apply: %s",
            ", ".join(s.get_name() for s in signal_types) or "None!",
        )

        indices: t.List[t.Tuple[t.Type[SignalType], SignalTypeIndex]] = []
        for s_type in signal_types:
            index = settings.index.load(s_type)
            if index is None:
                logging.info("No index for %s, skipping", s_type.get_name())
                continue
            indices.append((s_type, index))

        if not indices:
            self.stderr("No data to match against")
            return

        for path in self.files:
            for s_type, index in indices:
                seen = set()  # TODO - maybe take the highest certainty?
                if self.as_hashes:
                    results = _match_hashes(path, s_type, index)
                else:
                    results = _match_file(path, s_type, index)

                for r in results:
                    metadatas: t.List[t.Tuple[str, FetchedSignalMetadata]] = r.metadata
                    for collab, fetched_data in metadatas:
                        if not self.all and collab in seen:
                            continue
                        seen.add(collab)
                        # Supposed to be without whitespace, but let's make sure
                        distance_str = "".join(r.similarity_info.pretty_str().split())
                        print(
                            s_type.get_name(),
                            distance_str,
                            f"({collab})",
                            fetched_data,
                        )


def _match_file(
    path: pathlib.Path, s_type: t.Type[SignalType], index: SignalTypeIndex
) -> t.Sequence[IndexMatch]:
    if issubclass(s_type, MatchesStr):
        return index.query(path.read_text())
    assert issubclass(s_type, FileHasher)
    return index.query(s_type.hash_from_file(path))


def _match_hashes(
    path: pathlib.Path, s_type: t.Type[SignalType], index: SignalTypeIndex
) -> t.Sequence[IndexMatch]:
    ret: t.List[IndexMatch] = []
    for hash in path.read_text().splitlines():
        hash = hash.strip()
        if not hash:
            continue
        try:
            hash = s_type.validate_signal_str(hash)
        except Exception:
            logging.exception("%s failed verification on %s", s_type.get_name(), hash)
            hash_repr = repr(hash)
            if len(hash_repr) > 50:
                hash_repr = hash_repr[:47] + "..."
            raise CommandError(
                f"{hash_repr} from {path} is not a valid hash for {s_type.get_name()}",
                2,
            )
        ret.extend(index.query(hash))
    return ret



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

from pathlib import Path
from os import path
from setuptools import find_packages, setup

here = Path(__file__).parent

description = (here / "DESCRIPTION.rst").read_text(encoding="utf-8").strip()
long_description = (here / "README.md").read_text(encoding="utf-8").strip()
version = (here / "version.txt").read_text(encoding="utf-8").strip()

extensions_dir = here / "threatexchange" / "extensions"

extras_require = {}

for extension_dir in extensions_dir.iterdir():
    requirements = extension_dir / "requirements.txt"
    if requirements.is_file():
        extras_require[f"extensions.{extension_dir.name}"] = (
            requirements.read_text().strip().split("\n")
        )

all_extras = set(sum(extras_require.values(), []))
# We might not get any value from splitting these all out
extras_require["test"] = sorted({"pytest"} | all_extras)
extras_require["package"] = ["wheel"]
extras_require["lint"] = ["black"]
extras_require["types"] = ["mypy", "types-python-dateutil", "types-requests"]
extras_require["all"] = sorted(set(sum(extras_require.values(), [])))
# If you are developing pytx, use this install
# Note that without ffmpeg (for vpdq) you may get errors still
extras_require["dev"] = extras_require["all"]

setup(
    name="threatexchange",
    version=version,
    description="Python Library for Signal Exchange",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Facebook",
    author_email="threatexchange@fb.com",
    license="BSD",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3.8",
    ],
    keywords="facebook threatexchange ncmec stopncii pdq",
    url="https://www.github.com/facebook/ThreatExchange",
    packages=find_packages(exclude=["tests*"]),
    install_requires=[
        "python-Levenshtein",
        "requests>=2.26.0",
        "urllib3>=1.26.0",  # For allow_methods
        "python-dateutil",
        "dacite",
        "Pillow",  # pdq
        "pdqhash>=0.2.2",  # pdq
        "faiss-cpu>=1.6.3",  # faiss
        "numpy>=1.23.2",  # faiss
    ],
    extras_require=extras_require,
    entry_points={"console_scripts": ["threatexchange = threatexchange.cli.main:main"]},
)



# Copyright (c) Facebook, Inc. and its affiliates. All Rights Reserved

"""
Simple implementation for the NCMEC hash sharing XML API

You can find the complete documentation at 
https://report.cybertip.org/hashsharing/v2/documentation.pdf
"""

import xml.etree.ElementTree as ET
from datetime import datetime, timezone
from dataclasses import dataclass
from enum import Enum, unique
import logging
import time
import typing as t
from io import BytesIO
import html
import urllib.parse

import requests
from requests.packages.urllib3.util.retry import Retry

# Maybe move to a common library someday
from threatexchange.exchanges.clients.fb_threatexchange.api import TimeoutHTTPAdapter


_DATE_FORMAT_STR = "%Y-%m-%dT%H:%M:%SZ"
_DEFAULT_ELE = ET.Element("")

T = t.TypeVar("T")


def nullthrows(v: t.Optional[T]) -> T:
    """Typing helper"""
    assert v is not None
    return v


class _XMLWrapper:
    """
    Simpler wrapper around XML element for typing and null checking.

    To use the raw methods, just use .element
    """

    def __init__(self, e: ET.Element) -> None:
        self.element = e

    def __getitem__(self, key: str) -> "_XMLWrapper":
        """Asserts child with name exists"""
        child = self.element.findall(key)
        if not child:
            raise IndexError
        if len(child) != 1:
            raise ValueError(f"More than one child named {key}")
        return _XMLWrapper(child[0])

    def maybe(self, key: str, *rest: str) -> "_XMLWrapper":
        """Gets a child node or a placeholder for chaining"""
        curr = self.element
        for k in (key,) + rest:
            child = curr.find(k)
            if child is None:
                return _XMLWrapper(_DEFAULT_ELE)
            curr = child
        return _XMLWrapper(curr)

    @property
    def text(self) -> str:
        """Assert non-null text"""
        return nullthrows(self.element.text)

    @property
    def tag(self) -> str:
        """Suger for getting the tag name"""
        return self.element.tag

    def __len__(self) -> int:
        return len(self.element)

    def __bool__(self) -> bool:
        """True if not a non-existant element from .maybe()"""
        return self.element is not _DEFAULT_ELE

    def __iter__(self) -> t.Iterator["_XMLWrapper"]:
        """Iterate over children"""
        for child in self.element:
            yield _XMLWrapper(child)

    def __str__(self) -> str:
        return f"<{self.tag}>{self.element.text or '...'}</{self.tag}>"

    # Because these methods are named after types it messes
    # with typechecking
    def int(self, key: str) -> int:
        """Assert property exists and return as int"""
        return int(nullthrows(self.element.get(key)))

    def str(self, key: str) -> str:
        """Assert property exists and return as str"""
        return nullthrows(self.element.get(key))


@dataclass
class StatusResult:
    esp_id: int
    esp_name: str


@unique
class NCMECEntryType(Enum):
    image = "image"
    video = "video"


@dataclass
class NCMECEntryUpdate:
    id: str
    member_id: int
    entry_type: NCMECEntryType
    deleted: bool
    classification: t.Optional[str]
    fingerprints: t.Dict[str, str]

    @classmethod
    def from_xml(cls, xml: _XMLWrapper) -> "NCMECEntryUpdate":
        type_str = xml.tag.lower()
        deleted = type_str.startswith("deleted")
        if deleted:
            type_str = type_str[len("deleted") :]

        return cls(
            id=xml["id"].text,
            member_id=xml["member"].int("id"),
            entry_type=NCMECEntryType(type_str),
            deleted=deleted,
            classification=xml.maybe("classification").element.text,
            fingerprints={x.tag: x.text for x in xml.maybe("fingerprints")},
        )


class GetEntriesNextInfo(t.NamedTuple):
    """Wrapper around parsed "next" param data. Fields match param names"""

    start: int  # Might be- the start ID of the next fetch
    size: int  # How many records to fetch (always seems to be 1000)
    max: int  # The largest id in the time range

    @classmethod
    def from_next(cls, next_: str) -> t.Optional["GetEntriesNextInfo"]:
        s = html.unescape(next_)
        parsed = urllib.parse.parse_qs(s)
        start = parsed.get("start")
        size = parsed.get("size")
        max_ = parsed.get("max")
        if not start or not size or not max_:
            return None
        return GetEntriesNextInfo(int(start[0]), int(size[0]), int(max_[0]))


@dataclass
class GetEntriesResponse:
    updates: t.List[NCMECEntryUpdate]
    max_timestamp: int
    next: str

    @classmethod
    def from_xml(cls, xml: _XMLWrapper, fallback_max_time: int) -> "GetEntriesResponse":
        updates: t.List[NCMECEntryUpdate] = []
        max_ts = 0

        for content_xml in (xml.maybe("images"), xml.maybe("videos")):
            if not content_xml or not len(content_xml):
                continue
            max_ts = max(
                max_ts,
                int(
                    datetime.strptime(content_xml.str("maxTimestamp"), _DATE_FORMAT_STR)
                    .replace(tzinfo=timezone.utc)
                    .timestamp()
                ),
            )
            updates.extend(NCMECEntryUpdate.from_xml(c) for c in content_xml)

        next_ = xml.maybe("paging", "next").element.text or ""
        return cls(updates, max_ts or fallback_max_time, next_)

    def get_next_info(self) -> t.Optional[GetEntriesNextInfo]:
        if not self.next:
            return None
        return GetEntriesNextInfo.from_next(self.next)

    @property
    def estimated_entries_in_range(self) -> int:
        """Uses the "next" params to try and guess entries in range"""
        if not self.next:
            return len(self.updates)
        info = self.get_next_info()
        return (
            len(self.updates) + 1  # Parse error of some kind
            if info is None
            else info.max - info.start + len(self.updates)
        )


@unique
class NCMECEndpoint(Enum):
    status = "status"
    entries = "entries"
    members = "members"


class NCMECEnvironment(Enum):
    """
    Which 'Environment' to connect to.

    Environments differ by who is allowed to read or write, as well as what
    content exists on them.
    """

    Industry = "https://report.cybertip.org/hashsharing"
    NGO = "https://hashsharing.ncmec.org/npo"
    Exploitative = "https://hashsharing.ncmec.org/exploitative"

    test_Industry = "https://exttest.cybertip.org/hashsharing"
    test_NGO = "https://hashsharing-test.ncmec.org/npo"
    test_Exploitative = "https://hashsharing-test.ncmec.org/exploitative"


class NCMECHashAPI:
    """
    A wrapper around the NCMEC hash exchange API.

    Not to be confused with the cybertips API, which is for submitting illegal
    content, the hash exchange API is for submitting signals (or 'fingerprints'
    in the NCMEC parlance).

    Depending on which API you are connecting to, what is inside the database
    may differ greatly in exact topic. See NCMECEnvironment as well as NCMEC's
    documentation.
    """

    VERSION: t.ClassVar[str] = "v2"

    ENTRIES_PER_FETCH: t.ClassVar[int] = 1000

    def __init__(
        self,
        username: str,
        password: str,
        environment: NCMECEnvironment,
    ) -> None:
        assert is_valid_user_pass(username, password)
        self.username = username
        self.password = password
        self._base_url = environment.value

    def _get_session(self) -> requests.Session:
        """
        Custom requests sesson

        Ideally, should be used within a context manager:
        ```
        with self._get_session() as session:
            session.get()...
        ```
        """

        session = requests.Session()
        session.auth = (self.username, self.password)
        session.mount(
            self._base_url,
            adapter=TimeoutHTTPAdapter(
                timeout=60,
                max_retries=Retry(
                    total=4,
                    status_forcelist=[429, 500, 502, 503, 504],
                    # No retry for post. Could probably add timeout...
                    allowed_methods=["HEAD", "GET", "OPTIONS"],
                    backoff_factor=0.2,  # ~1.5 seconds of retries
                ),
            ),
        )
        return session

    def _get(self, endpoint: NCMECEndpoint, *, next_: str = "", **params) -> ET.Element:
        """
        Perform an HTTP GET request, and return the XML response payload.

        Same timeouts and retry strategy as `_get_session` above.
        """

        url = "/".join((self._base_url, self.VERSION, endpoint.value))
        if next_:
            url = self._base_url + next_
            params = {}

        with self._get_session() as session:
            response = session.get(url, params=params)
        # Gate this log just in case decode() blows up
        if logging.getLogger().isEnabledFor(logging.DEBUG):
            logging.debug("GET %s returned: %s", endpoint, response.content.decode())
        response.raise_for_status()
        # Namespaces don't really add anything here, so let's remove them
        it = ET.iterparse(BytesIO(response.content))
        for _, el in it:
            _ns, has_namespace, postfix = el.tag.partition("}")
            if has_namespace:
                el.tag = postfix
        return it.root  # type: ignore

    def _post(self, endpoint: NCMECEndpoint, *, data=None) -> t.Any:
        """
        Perform an HTTP POST request, and return the XML response payload.

        No timeout or retry strategy.
        """

        url = "/".join((self._base_url, endpoint.value))
        with self._get_session() as session:
            response = session.post(url, data=data)
            response.raise_for_status()
            return response

    def status(self) -> StatusResult:
        """Query the status endpoint, which tells you who you are."""
        response = self._get(NCMECEndpoint.status)
        member = _XMLWrapper(response)["member"]
        return StatusResult(member.int("id"), member.text)

    def members(self) -> t.List[StatusResult]:
        """Query the members endpoint, which gives you a list of esps"""
        response = self._get(NCMECEndpoint.members)
        return [
            StatusResult(member.int("id"), member.text)
            for member in _XMLWrapper(response)
        ]

    def get_entries(
        self,
        *,
        start_timestamp: int = 0,
        end_timestamp: int = 0,
        next_: str = "",
    ) -> GetEntriesResponse:
        """
        Fetch a series of update records from the hash API.

        DANGER! The NCMEC API does not return entries in time order! If you
        don't exhaust the entire iterator, you can't trust the max timestamp!
        """
        end_timestamp = end_timestamp or int(time.time())
        # Need a dict here for python keyword 'from'
        params: t.Dict[str, t.Any] = {
            "from": _date_format(start_timestamp),
        }
        response = self._get(
            NCMECEndpoint.entries,
            next_=next_,
            to=_date_format(end_timestamp),
            **params,
        )
        return GetEntriesResponse.from_xml(_XMLWrapper(response), int(time.time()))

    def get_entries_iter(
        self, *, start_timestamp: int = 0, end_timestamp: int = 0
    ) -> t.Iterator[GetEntriesResponse]:
        """
        A simple wrapper around get_entries to keep fetching until complete.

        If you don't exhaust the iterator, you can't make any assumptions about how
        much of the data you have fetched. @see get_entries
        """
        has_more = True
        next_ = ""
        while has_more:
            result = self.get_entries(
                start_timestamp=start_timestamp,
                end_timestamp=end_timestamp,
                next_=next_,
            )
            next_ = result.next
            has_more = bool(next_)
            yield result


def _date_format(timestamp: int) -> str:
    """ISO 8601 format yyyy-MM-dd'T'HH:mm:ss.SSSZ"""
    return datetime.fromtimestamp(timestamp).strftime(_DATE_FORMAT_STR)


def is_valid_user_pass(user: str, password: str) -> bool:
    return bool(user and password)  # Is there anything more we can do here?
